## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.wrappers
## -- Module  : optuna.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2022-03-24  0.0.0     SY       Creation 
## -- 2022-03-24  1.0.0     SY       Release of first version
## -- 2022-03-25  1.0.1     SY       Change methods names: _ofct_optuna and get_parameters
## -- 2022-04-05  1.0.2     SY       Add tuning recap visualization: class _plot_results
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.2 (2022-04-05)

This module provides a wrapper class for hyperparameter tuning by reusing Optuna framework
"""


import optuna
from mlpro.bf.ml import *
from mlpro.bf.math import *
from mlpro.bf.various import *
from mlpro.rl.models import *
from mlpro.gt.models import *
import os




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class WrHPTOptuna(HyperParamTuner, ScientificObject):
    """
    This class is a ready to use wrapper class for Optuna framework. 
    Objects of this type can be treated as a hyperparameter tuner object.
    
    Parameters
    ----------
    p_logging: Log
        Log level (see constants for log levels)
    p_ids : list of str, optional
        List of hyperparameter ids to be tuned, otherwise all hyperparameters, default: None
    p_visualization : boolean
        enable visualization at the end of the tuning, default: False
        
    Attributes
    ----------
    C_NAME: str
        Name of the class.
    """
    
    C_NAME              = 'Optuna'
        
    C_SCIREF_TYPE       = ScientificObject.C_SCIREF_TYPE_PROCEEDINGS
    C_SCIREF_AUTHOR     = "Akiba, Takuya and Sano, Shotaro and Yanase, Toshihiko and Ohta, Takeru and Koyama, Masanori"
    C_SCIREF_TITLE      = "Optuna: A Next-Generation Hyperparameter Optimization Framework"
    C_SCIREF_YEAR       = "2019"
    C_SCIREF_ISBN       = "9781450362016"
    C_SCIREF_PUBLISHER  = "Association for Computing Machinery"
    C_SCIREF_CITY       = "New York"
    C_SCIREF_COUNTRY    = "USA"
    C_SCIREF_URL        = "https://doi.org/10.1145/3292500.3330701"
    C_SCIREF_DOI        = "10.1145/3292500.3330701"
    C_SCIREF_PAGES      = "2623–2631"
    C_SCIREF_BOOKTITLE  = "Proceedings of the 25th ACM SIGKDD International Conference on Knowledge Discovery & Data Mining"
    
    C_LOG_SEPARATOR = '------------------------------------------------------------------------------'
    

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_logging=Log.C_LOG_ALL, p_ids=None, p_visualization=False):
        super().__init__(p_logging=p_logging)

        self._ids = p_ids
        self.num_trials = 0
        self.visualize = p_visualization
        
        self.log(self.C_LOG_TYPE_I, 'Optuna configuration is successful')


## -------------------------------------------------------------------------------------------------
    def _maximize(self) -> TrainingResults:
        """
        This method is a place to setup a hp tuner based on hp structure of the model
        and run the hp tuner. Create a study object and execute the optimization.

        Returns
        -------
        best_result : float
            The best result after a number of evaluations.

        """
        
        if self._training_cls is None:
            raise ParamError('Mandatory parameter self._training_cls is not supplied')
        
        if self._num_trials <= 0:
            raise ParamError('Parameter self._num_trials must be greater than 0')
        
        if self._root_path is None:
            raise ParamError('Mandatory parameter self._root_path is not supplied')
        
        if self._training_param is None:
            raise ParamError('Mandatory parameter self._training_param is not supplied')
        
        # change root path in training param
        self._training_param['p_training_param']['p_path'] = self._root_path+os.sep+'HyperparameterTuning'+os.sep+'Base (Preparation)'
        if not os.path.exists(self._training_param['p_training_param']['p_path']):
            os.mkdir(self._root_path+os.sep+'HyperparameterTuning')
            os.mkdir(self._training_param['p_training_param']['p_path'])
        
        # ignore collecting data during tuning to save tuning time and memory
        self._training_param['p_training_param']['p_collect_states'] = False
        self._training_param['p_training_param']['p_collect_actions'] = False
        self._training_param['p_training_param']['p_collect_rewards'] = False
        self._training_param['p_training_param']['p_logging'] = Log.C_LOG_NOTHING
        self._training_param['p_training_param']['p_visualize'] = False
        self._training_param['p_training_param']['p_collect_eval'] = True
        
        # instantiate a scenario class and define the model in a variable
        training_cls = self._training_cls(**self._training_param['p_training_param'])
        self._model = training_cls._scenario._model
        
        # prepare a data storing class
        for x, _id in enumerate(self._model._hyperparam_tuple.get_dim_ids()):
            hp_object = self._model._hyperparam_tuple.get_related_set().get_dim(_id)
            self.variables.append(hp_object.get_name_short()+'_'+str(x))
        self.HPDataStoring = DataStoring(self.variables)
        self.HPDataStoring.add_frame('HP_0')
    
        # run the trials and gain the highest score
        study = optuna.create_study(direction="maximize")
        study.optimize(self._ofct_optuna, n_trials=self._num_trials)
        best_trial = study.best_trial
        best_result = best_trial.value
        best_param = study.best_params
        self.save(best_param, best_result, 'best_parameters.csv')
        
        if self.visualize:
            self._plot_results(study)
        
        return best_result


## -------------------------------------------------------------------------------------------------
    def _ofct_optuna(self, trial):
        """
        Wrap model training with an objective function and return the output score.

        Parameters
        ----------
        trial : object
            Suggest hyperparameters using a trial object.

        Returns
        -------
        result.highscore : float
            final score of a trial.

        """
        self.log(self.C_LOG_TYPE_I, 'Trial number '+str(self.num_trials)+' has started')
        self.log(self.C_LOG_TYPE_I, self.C_LOG_SEPARATOR, '\n')

        # change root path in training param
        self._training_param['p_training_param']['p_path'] =  self._root_path+os.sep+'HyperparameterTuning'+os.sep+'Trial_'+str(self.num_trials)
        if not os.path.exists(self._training_param['p_training_param']['p_path']):
            os.mkdir(self._training_param['p_training_param']['p_path'])

        # instantiate a scenario class
        training_cls = self._training_cls(**self._training_param['p_training_param'])
        self._model = training_cls._scenario._model
        
        # setup parameters that compatible to optuna spaces
        p_params = self.get_parameters(trial)
        for x, _id in enumerate(self._model._hyperparam_tuple.get_dim_ids()):
            self._model._hyperparam_tuple.set_value(_id, p_params[x])

        # run the scenario and retrieve the high score
        result = training_cls.run()

        # store trial parameters and the trial's result
        self.HPDataStoring.memorize('Trial', 'HP_0', self.num_trials)
        self.HPDataStoring.memorize('Highscore', 'HP_0', result.highscore)
        for x, _id in enumerate(self._model._hyperparam_tuple.get_dim_ids()):
            hp_name_short = self._model._hyperparam_tuple.get_related_set().get_dim(_id).get_name_short()
            self.HPDataStoring.memorize(hp_name_short+'_'+str(x), 'HP_0', p_params[x])

        self.num_trials += 1

        self.log(self.C_LOG_TYPE_I, 'Trial number '+str(self.num_trials)+' has finished')
        self.log(self.C_LOG_TYPE_I, self.C_LOG_SEPARATOR, '\n')
        
        return result.highscore
        

## -------------------------------------------------------------------------------------------------
    def get_parameters(self, trial):
        """
        This method is used to get parameters within boundaries.
        The hyperparameter should be bounded both above and below.
        For different parameter expressions, please redefined this method and check https://optuna.readthedocs.io/en/stable/reference/generated/optuna.trial.Trial.html!
        For big data handling, please redifined this method!

        Parameters
        ----------
        trial : object
            Suggest hyperparameters using a trial object.
        
        Returns
        -------
        parameters : list
            List of parameter expressions.

        """
        
        if self._model._hyperparam_tuple is None:
            self._model._init_hyperparam()
        
        parameters = []
        for x, _id in enumerate(self._model._hyperparam_tuple.get_dim_ids()):
            hp_object = self._model._hyperparam_tuple.get_related_set().get_dim(_id)
            hp_boundaries = hp_object.get_boundaries()
            if not hp_boundaries:
                raise ImplementationError('Missing boundary of a hyperparameter!')
            else:
                hp_low = hp_boundaries[0]
                hp_high = hp_boundaries[1]
                if hp_object._base_set == Dimension.C_BASE_SET_N or hp_object._base_set == Dimension.C_BASE_SET_Z:
                    parameters.append(trial.suggest_int(hp_object.get_name_short()+'_'+str(x),hp_low,hp_high))
                elif hp_object._base_set == Dimension.C_BASE_SET_R:
                    parameters.append(trial.suggest_uniform(hp_object.get_name_short()+'_'+str(x),hp_low,hp_high))
                else:
                    raise ImplementationError('Missing a short name of a hyperparameter!')
        
        self.log(self.C_LOG_TYPE_I, 'New parameters for optuna tuner is ready')
        self.log(self.C_LOG_TYPE_I, self.C_LOG_SEPARATOR, '\n')
        
        return parameters


## -------------------------------------------------------------------------------------------------
    def _plot_results(self, p_study):
        """
        Visualize the tuning recap.

        Parameters
        ----------
        p_study : Study
            Optuna Study object.

        """
        self.log(self.C_LOG_TYPE_I, 'Plotting of tuning recap is started')
        
        fig = optuna.visualization.plot_slice(p_study)
        fig.show()
        
        fig = optuna.visualization.plot_param_importances(p_study)
        fig.show()
        
        fig = optuna.visualization.plot_parallel_coordinate(p_study)
        fig.show()
        
        self.log(self.C_LOG_TYPE_I, 'Plotting of tuning recap is succesful')



## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro
## -- Module  : test_various
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-04-16  1.0.0     DA       Creation
## -- 2021-09-11  1.0.0     MRD      Change Header information to match our new library name
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.0 (2021-04-16)

Unit test classes for various basic functions.
"""


import pytest
from mlpro.bf.various import Log



## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class MyLog(Log):
    C_TYPE      = 'Test class'
    C_NAME      = 'MyLog'

    def __init__(self): 
        Log.__init__(self)


## -------------------------------------------------------------------------------------------------
def test_logging(capsys):
    """
    Method description.
    """

    lo_log = MyLog()
    lo_log.log(Log.C_LOG_TYPE_I, 'Information')
    lo_log.log(Log.C_LOG_TYPE_W, 'Warning')
    lo_log.log(Log.C_LOG_TYPE_E, 'Error')
    captured = capsys.readouterr()
    assert 'Information' in captured.out
    assert 'Warning' in captured.out
    assert 'Error' in captured.out
    lo_log.switch_logging(Log.C_LOG_NOTHING)
    lo_log.log(Log.C_LOG_TYPE_I, 'Information')
    lo_log.log(Log.C_LOG_TYPE_W, 'Warning')
    lo_log.log(Log.C_LOG_TYPE_E, 'Error')
    captured = capsys.readouterr()
    assert 'Information' not in captured.out
    assert 'Warning' not in captured.out
    assert 'Error' not in captured.out
    lo_log.switch_logging(Log.C_LOG_WE)
    lo_log.log(Log.C_LOG_TYPE_I, 'Information')
    lo_log.log(Log.C_LOG_TYPE_W, 'Warning')
    lo_log.log(Log.C_LOG_TYPE_E, 'Error')
    captured = capsys.readouterr()
    assert 'Information' not in captured.out
    assert 'Warning' in captured.out
    assert 'Error' in captured.out
    lo_log.switch_logging(Log.C_LOG_E)
    lo_log.log(Log.C_LOG_TYPE_I, 'Information')
    lo_log.log(Log.C_LOG_TYPE_W, 'Warning')
    lo_log.log(Log.C_LOG_TYPE_E, 'Error')
    captured = capsys.readouterr()
    assert 'Information' not in captured.out
    assert 'Warning' not in captured.out
    assert 'Error' in captured.out
    
        



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl
## -- Module  : sciui_rl.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-07-30  0.0.0     SY       Creation
## -- 2021-10-07  1.0.0     SY       Release of first draft
## -- 2022-01-06  1.0.1     DA       Refactoring
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.1 (2022-01-06)

This file provides reusable SciUI components for SciUI_RL.
"""



from tkinter import *
import tkinter
import tkinter.font as tkFont
from tkinter.ttk import *
from tkinter.scrolledtext import *
from mpl_toolkits.mplot3d import Axes3D

from mlpro.bf.ui.sciui.framework import *
from mlpro.bf.math import ESpace
import mlpro.rl.pool.scenarios as ScePool
from mlpro.rl.models import *
import os
import pkgutil
import numpy as np



## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLDetails(SciUIFrameParam):
    """
    ...
    """

    C_NAME          = 'Details'
    C_SCE_LISTS     = []

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh  = True
        self.sce_var        = StringVar()
        self.env_var        = StringVar()
        self.alg_var        = StringVar()
        
        self.auto_scan_sce  = Scenario.__subclasses__()
        for i in range(len(self.auto_scan_sce)):
            if self.auto_scan_sce[i].C_NAME != '????':
                self.C_SCE_LISTS.append(self.auto_scan_sce[i].C_NAME)  


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        if self.first_refresh: 
            self.first_refresh = False
        else: 
            if self.shared_db.rl_sce_state_btn == False:
                self.button_sce.config(state="disabled")
                

        super().refresh(p_parent_frame=p_parent_frame)
        
        self.sce_label      = Label(self.frame, text='Scenario', pady=3, font=self.fontstyle, width=20, height=1, anchor='w').grid(row=0, column=0, sticky='W')
        self.sce_box        = Entry(self.frame, textvariable=self.sce_var, width=28, justify='left', state='disabled', background='white', font=self.fontstyle)
        self.sce_box.grid(row=0, column=1)
        self.sce_var.trace('w', self.__cb_sce_changed)
        self.sce_button     = Frame(self.frame, highlightbackground = "black", bd=0, width=7)
        self.sce_button.grid(row=0, column=2, sticky='w')
        self.button_sce     = Button(self.sce_button, text='...', font=self.fontstyle, fg = 'black', command=self.__cb_sce_button)
        self.button_sce.pack()
        self.sce_i_button   = Frame(self.frame, highlightbackground = "black", bd=0, width=5)
        self.sce_i_button.grid(row=0, column=3, sticky='w')
        self.button_sce_i   = Button(self.sce_i_button, text='i', font=self.fontstyle, fg = 'black', state='disabled', command=self.__cb_sce_button_info)
        self.button_sce_i.pack()
        
        self.env_label      = Label(self.frame, text='Environment', pady=3, font=self.fontstyle, width=20, height=1, anchor='w').grid(row=1, column=0, sticky='W')
        self.env_box        = Entry(self.frame, textvariable=self.env_var, width=28, justify='left', state='disabled', background='white', font=self.fontstyle)
        self.env_box.grid(row=1, column=1)
        self.env_var.trace('w', self.__cb_env_changed)
        
        self.alg_label      = Label(self.frame, text='Algorithm', pady=3, font=self.fontstyle, width=20, height=1, anchor='w').grid(row=2, column=0, sticky='W')
        self.alg_box        = Entry(self.frame, textvariable=self.alg_var, width=28, justify='left', state='disabled', background='white', font=self.fontstyle)
        self.alg_box.grid(row=2, column=1)
        self.alg_var.trace('w', self.__cb_alg_changed)
        

## -------------------------------------------------------------------------------------------------
    def __cb_sce_changed(self, *args):
        try:
            sce   = self.sce_var.get()
        except:
            return

        self.sce_var.set(sce)
        self.env_var.set(self.shared_db.rl_env)
        self.alg_var.set(self.shared_db.rl_learning_alg)
        
        self.shared_db.rlui_selected_scenario = True
        self.shared_db.start_global_refresh()
        

## -------------------------------------------------------------------------------------------------
    def __cb_env_changed(self, *args):
        pass
    
    
## -------------------------------------------------------------------------------------------------
    def __cb_alg_changed(self, *args):
        pass
    

## -------------------------------------------------------------------------------------------------
    def __cb_sce_button(self, *args):
        self.button_sce.config(state="disabled")
        
        self.sce_pop                = Toplevel(self.frame)
        self.sce_pop.title("RL Scenario Pool")
        self.sce_pop.geometry("500x450")
        
        self.left_frame_sce         = Frame(self.sce_pop)
        self.left_frame_sce.pack(side='left', fill='both')
        
        self.sce_list               = Listbox(self.left_frame_sce, width=72, height=24)
        self.sce_list.grid(row=0, column=0, sticky='w')
        for item in self.C_SCE_LISTS:
            self.sce_list.insert('end',item)
        self.sce_list.pack()
        
        self.sce_scrollbar          = Scrollbar(self.sce_pop)
        self.sce_scrollbar.pack(side='right', fill='both')
        self.sce_list.config(yscrollcommand = self.sce_scrollbar.set)
        self.sce_scrollbar.config(command = self.sce_list.yview)
        
        self.apply_button_sce       = Button(self.left_frame_sce, text='Apply', font=self.fontstyle, fg = 'black', command=self.__cb_apply_scenario)
        self.apply_button_sce.pack()
        
        self.cancel_button_sce      = Button(self.left_frame_sce, text='Cancel', font=self.fontstyle, fg = 'black', command=self.__cb_cancel_scenario)
        self.cancel_button_sce.pack()
        
        self.sce_pop.protocol("WM_DELETE_WINDOW", self.__cb_sce_pop_closing)
    

## -------------------------------------------------------------------------------------------------
    def __cb_sce_pop_closing(self, *args):
        self.sce_pop.destroy()
        self.button_sce.config(state="normal")

## -------------------------------------------------------------------------------------------------
    def __cb_sce_button_info(self, *args):
        self.sce_i_pop       = Toplevel(self.frame)
        self.sce_i_pop.title("Information about Environment")
        self.sce_i_pop.geometry("500x450")
        self.sce_i_label      = Label(self.sce_i_pop, text='No information!', pady=3, width=20, height=1, anchor='w').grid(row=0, column=0, sticky='W')
        
        
## -------------------------------------------------------------------------------------------------
    def __cb_apply_scenario(self, *args):
        try:
            self.applied_sce                = self.sce_list.get('active')
        except:
            return
        
        for i in range(len(self.auto_scan_sce)):
            if self.auto_scan_sce[i].C_NAME == self.applied_sce:
                self.shared_db.rl_scenario          = self.auto_scan_sce[i](p_mode=Environment.C_MODE_SIM, p_ada=True, p_cycle_limit=10, p_visualize=True, p_logging=True)
                self.shared_db.rl_scenario_class    = self.auto_scan_sce[i]
                break
        self.shared_db.rl_env               = self.shared_db.rl_scenario.get_env().C_NAME
        self.shared_db.rl_learning_alg      = self.shared_db.rl_scenario.get_agent().C_NAME
        
        self.sce_var.set(self.applied_sce)
        self.sce_pop.destroy()
        self.button_sce.config(state="normal")
        self.button_sce_i.config(state="normal")
        
    
        
## -------------------------------------------------------------------------------------------------
    def __cb_cancel_scenario(self, *args):
        self.sce_pop.destroy()
        self.button_sce.config(state="normal")
    
    
    
    
        
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLParam(SciUITabCTRL):
    """
    ...
    """

    C_NAME          = 'Parameters'
    
    C_FONT_FAMILY   = 'Lucida Grande'
    C_FONT_SIZE     = 10 

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh              = True
        self.fontstyle                  = tkFont.Font(family=self.C_FONT_FAMILY, size=self.C_FONT_SIZE)
        self.tab_list                   = []
        self.tab_ctrl                   = None 
        self.add_tab('Environment', None)
        self.add_tab('RL Agent(s)', None)
        self.add_tab('Data Storing', None)
        
        self.ep_var                     = IntVar()
        self.cycle_time_var             = IntVar()
        self.ep_var.set(10)
        self.cycle_time_var.set(100)
        
        self.param_var                  = []
        
        self.save_data                  = IntVar()
        self.collect_states             = IntVar()
        self.collect_actions            = IntVar()
        self.collect_rewards            = IntVar()
        self.collect_training           = IntVar()
        self.data_path                  = StringVar()
        self.save_data.set(self.shared_db.rl_save_data)
        self.collect_states.set(self.shared_db.rl_collect_states)
        self.collect_actions.set(self.shared_db.rl_collect_actions)
        self.collect_rewards.set(self.shared_db.rl_collect_rewards)
        self.collect_training.set(self.shared_db.rl_collect_training)
        self.data_path.set("%s"%os.getcwd())

## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        if not self.first_refresh: 
            self.__del__()
            self.init_component()
        self.first_refresh = False
        super().refresh(p_parent_frame=p_parent_frame)
        self.tab_ctrl.config(width=400, height=572)
        self.__cb_frame()


 ## -------------------------------------------------------------------------------------------------
    def __cb_frame(self, *args): 
            
        ### Scrollbar ###
        env_canvas          = Canvas(self.tab_list[0][1])
        env_canvas.pack(side='left', expand=1, fill ="both")
        env_scrollbar       = Scrollbar(self.tab_list[0][1], orient="vertical", command=env_canvas.yview)
        env_scrollbar.pack(side='right', fill='y')
        env_canvas.configure(yscrollcommand=env_scrollbar.set)
        env_canvas.bind('<Configure>', lambda e:env_canvas.configure(scrollregion=env_canvas.bbox("all")))
        self.env_frame      = Frame(env_canvas)
        env_canvas.create_window((0,0), window=self.env_frame, anchor="nw")
        
        alg_canvas          = Canvas(self.tab_list[1][1])
        alg_canvas.pack(side='left', expand=1, fill ="both")
        alg_scrollbar       = Scrollbar(self.tab_list[1][1], orient="vertical", command=alg_canvas.yview)
        alg_scrollbar.pack(side='right', fill='y')
        alg_canvas.configure(yscrollcommand=alg_scrollbar.set)
        alg_canvas.bind('<Configure>', lambda e:alg_canvas.configure(scrollregion=alg_canvas.bbox("all")))
        self.alg_frame      = Frame(alg_canvas)
        alg_canvas.create_window((0,0), window=self.alg_frame, anchor="nw")
        
        data_canvas         = Canvas(self.tab_list[2][1])
        data_canvas.pack(side='left', expand=1, fill ="both")
        data_scrollbar      = Scrollbar(self.tab_list[2][1], orient="vertical", command=data_canvas.yview)
        data_scrollbar.pack(side='right', fill='y')
        data_canvas.configure(yscrollcommand=data_scrollbar.set)
        data_canvas.bind('<Configure>', lambda e:data_canvas.configure(scrollregion=data_canvas.bbox("all")))
        self.data_frame     = Frame(data_canvas)
        data_canvas.create_window((0,0), window=self.data_frame, anchor="nw")
        
        if (self.shared_db.rlui_selected_scenario == 0):
            Label(self.env_frame, text ="Please follow the given procedures :").grid(row=0, column=0, padx = 5, pady = 0, sticky="w")
            Label(self.env_frame, text ="1. Select RL scenario").grid(row=1, column=0, padx = 5, pady = 0, sticky="w")
            Label(self.env_frame, text ="3. Fill all the parameters").grid(row=2, column=0, padx = 5, pady = 0, sticky="w")
            Label(self.env_frame, text ="4. Run environment").grid(row=3, column=0, padx = 5, pady = 0, sticky="w")
        else:
            self.__cb_environment_tab()
            self.__cb_agents_tab()
            self.__cb_data_storing_tab()
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_environment_tab(self, *args):
        Label(self.env_frame, text='Max. Episodes', pady=3, font=self.fontstyle).grid(row=0, column=0, padx=5, pady=0, sticky='NW')
        self.ep_inputs = Entry(self.env_frame, textvariable=self.ep_var, width=5, justify='center', font=self.fontstyle, state=self.shared_db.rl_param_state_box)
        self.ep_inputs.grid(row=0, column=1, sticky='w')
        
        Label(self.env_frame, text='Cycle Time / Ep.', pady=3, font=self.fontstyle).grid(row=1, column=0, padx=5, pady=0, sticky='NW')
        self.cycle_time_inputs = Entry(self.env_frame, textvariable=self.cycle_time_var, width=10, justify='center', font=self.fontstyle, state=self.shared_db.rl_param_state_box)
        self.cycle_time_inputs.grid(row=1, column=1, sticky='w')
        
        Label(self.env_frame, text='  ', pady=3, font=self.fontstyle).grid(row=2, column=0, padx=5, pady=0, sticky='NW')
        Label(self.env_frame, text='  ', pady=3, font=self.fontstyle).grid(row=3, column=0, padx=5, pady=0, sticky='NW')
        
        self.submit_cycle_border = Frame(self.env_frame, highlightbackground = "black", highlightthickness = 1, bd=0)
        self.submit_cycle_border.grid(row=4, column=1, sticky='w')
        self.submit_cycle = Button(self.submit_cycle_border, text='Run - Next Cycle Time', font=self.fontstyle, fg = 'black', bg = 'honeydew2', command=self.__cb_simulation_run_cycle, state=self.shared_db.rl_run_state_btn)
        self.submit_cycle.pack()
        
        self.submit_ep_border = Frame(self.env_frame, highlightbackground = "black", highlightthickness = 1, bd=0)
        self.submit_ep_border.grid(row=5, column=1, sticky='w')
        self.submit_ep = Button(self.submit_ep_border, text='Run - Next Episode', font=self.fontstyle, fg = 'black', bg = 'LightCyan2', command=self.__cb_simulation_run_ep, state=self.shared_db.rl_run_state_btn)
        self.submit_ep.pack()
        
        self.submit_border = Frame(self.env_frame, highlightbackground = "black", highlightthickness = 1, bd=0)
        self.submit_border.grid(row=6, column=1, sticky='w')
        self.submit = Button(self.submit_border, text='Run', font=self.fontstyle, fg = 'black', bg = 'LightSteelBlue1', command=self.__cb_simulation_run, state=self.shared_db.rl_run_state_btn)
        self.submit.pack()
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_agents_tab(self, *args):
        try:
            num_agents              = len(self.shared_db.rl_scenario.get_agent().get_agents())
            multiagent              = True
        except:
            num_agents = 1
            multiagent              = False
        
        n_row   = 0
        for n in range(num_agents):
            if multiagent:
                self.agent_         = self.shared_db.rl_scenario.get_agent().get_agents()[n][0]
            else:
                self.agent_         = self.shared_db.rl_scenario.get_agent()
            Label(self.alg_frame, text=self.agent_.get_name(), pady=3, font=self.fontstyle).grid(row=n_row, column=0, padx=5, pady=0, sticky='NW')
            n_row += 1
            for x in range(self.agent_._policy._hyperparam_space.get_num_dim()):
                param_ids           = self.agent_._policy._hyperparam_space.get_dim_ids()[x]
                param_name          = self.agent_._policy._hyperparam_space.get_dim(param_ids).get_name_short()
                param_baseset       = self.agent_._policy._hyperparam_space.get_dim(param_ids).get_base_set()
                param_val           = self.agent_._policy._hyperparam_tupel.get_value(param_ids)
                if param_baseset == 'N' or param_baseset == 'Z':
                    self.param_var.append(IntVar())
                else:
                    self.param_var.append(DoubleVar())
                Label(self.alg_frame, text=param_name, pady=3, font=self.fontstyle).grid(row=n_row, column=0, padx=5, pady=0, sticky='NW')
                self.param_var[-1].set(param_val)
                self.param_inputs   = Entry(self.alg_frame, textvariable=self.param_var[-1], width=15, justify='center', font=self.fontstyle, state=self.shared_db.rl_param_state_box)
                self.param_inputs.grid(row=n_row, column=1, sticky='w')
                n_row += 1
            Label(self.alg_frame, text=' ', pady=3, font=self.fontstyle).grid(row=n_row, column=0, padx=5, pady=0, sticky='NW')
            n_row += 1       
        
            
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_data_storing_tab(self, *args):
        
        Checkbutton(self.data_frame, text = "Save data", justify='left', variable=self.save_data, state=self.shared_db.rl_param_state_box).grid(row=0, column=0, padx=5, pady=0, sticky='NW')
        Checkbutton(self.data_frame, text = "Collect states", justify='left', variable=self.collect_states, state=self.shared_db.rl_param_state_box).grid(row=1, column=0, padx=5, pady=0, sticky='NW')
        Checkbutton(self.data_frame, text = "Collect actions", justify='left', variable=self.collect_actions, state=self.shared_db.rl_param_state_box).grid(row=2, column=0, padx=5, pady=0, sticky='NW')
        Checkbutton(self.data_frame, text = "Collect rewards", justify='left', variable=self.collect_rewards, state=self.shared_db.rl_param_state_box).grid(row=3, column=0, padx=5, pady=0, sticky='NW')
        Checkbutton(self.data_frame, text = "Collect training", justify='left', variable=self.collect_training, state=self.shared_db.rl_param_state_box).grid(row=4, column=0, padx=5, pady=0, sticky='NW')
        
        Label(self.data_frame, text='Path', padx=5, pady=0, font=self.fontstyle).grid(row=5, column=0, sticky='NW')
        self.path = Entry(self.data_frame, width=35, textvariable=self.data_path, justify='left', font=self.fontstyle)
        self.path.grid(row=5, column=1, sticky='w')
        
            
 ## -------------------------------------------------------------------------------------------------
    def __first_training_run(self, *args):
        if self.shared_db.rl_sim_started == False and self.shared_db.rl_sim_stop == True:
            cycle_time                  = self.cycle_time_var.get()
            self.shared_db.rl_scenario  = self.shared_db.rl_scenario_class(p_mode=Environment.C_MODE_SIM, p_ada=True, p_cycle_limit=cycle_time, p_visualize=True, p_logging=True)
            
            try:
                num_agents              = len(self.shared_db.rl_scenario.get_agent().get_agents())
                multiagent              = True
            except:
                num_agents = 1
                multiagent              = False
                
            n_param = 0
            for n in range(num_agents):
                if multiagent:
                    self.agent_         = self.shared_db.rl_scenario.get_agent().get_agents()[n][0]
                else:
                    self.agent_         = self.shared_db.rl_scenario.get_agent()
                for x in range(self.agent_._policy._hyperparam_space.get_num_dim()):
                    param_ids           = self.agent_._policy._hyperparam_space.get_dim_ids()[x]
                    param_value         = self.param_var[n_param].get()
                    self.agent_._policy._hyperparam_tupel.set_value(param_ids,param_value)
                    n_param += 1
                    
            self.shared_db.rl_save_data         = self.save_data.get()
            self.shared_db.rl_collect_states    = self.collect_states.get()
            self.shared_db.rl_collect_actions   = self.collect_actions.get()
            self.shared_db.rl_collect_rewards   = self.collect_rewards.get()
            self.shared_db.rl_collect_training  = self.collect_training.get()
            ep_limit                            = self.ep_var.get()
            self.shared_db.rl_training  = Training(p_scenario=self.shared_db.rl_scenario,
                                                   p_episode_limit=ep_limit,
                                                   p_cycle_limit=cycle_time,
                                                   p_collect_states=self.shared_db.rl_collect_states,
                                                   p_collect_actions=self.shared_db.rl_collect_actions,
                                                   p_collect_rewards=self.shared_db.rl_collect_rewards,
                                                   p_collect_training=self.shared_db.rl_collect_training,
                                                   p_logging=self.shared_db.rl_log_training
                                                   )
            self.shared_db.rl_scenario._env.switch_logging(self.shared_db.rl_log_env)
            self.shared_db.rl_scenario._agent.switch_logging(self.shared_db.rl_log_agent)
            self.shared_db.rl_scenario.switch_logging(self.shared_db.rl_log_process)
         
        
 ## -------------------------------------------------------------------------------------------------
    def __env_finish_check(self, *args):
        if self.shared_db.rl_training._episode_id >= self.shared_db.rl_training._episode_limit-1:
            self.shared_db.rl_sim_finished = True
         
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_simulation_run(self, *args):
        self.__first_training_run()
        self.shared_db.last_refresh         = 0
        if not self.shared_db.rl_sim_finished:
            self.shared_db.rl_sim_started   = True
            self.shared_db.rl_sim_stop      = False
            while self.shared_db.rl_training._episode_id < self.shared_db.rl_training._episode_limit:
                current_episode_id          = self.shared_db.rl_training._episode_id
                while self.shared_db.rl_training._episode_id == current_episode_id:
                    self.shared_db.rl_training.run_cycle()
                    if (self.shared_db.last_refresh % self.shared_db.refresh_rate) == 0:
                        self.shared_db.ep_monitor_frame.refresh(None)
                        self.shared_db.plot_frame.refresh(None)
                        self.shared_db.log_frame.refresh(None)
                    self.shared_db.last_refresh += 1
            self.shared_db.rl_sim_stop      = True
            self.__env_finish_check()
            self.shared_db.start_global_refresh()
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_simulation_run_ep(self, *args):
        self.__first_training_run()
        self.shared_db.last_refresh         = 0
        if not self.shared_db.rl_sim_finished:
            self.shared_db.rl_sim_started   = True
            self.shared_db.rl_sim_stop      = False
            current_episode_id              = self.shared_db.rl_training._episode_id
            while self.shared_db.rl_training._episode_id == current_episode_id:
                self.shared_db.rl_training.run_cycle()
                if (self.shared_db.last_refresh % self.shared_db.refresh_rate) == 0:
                    self.shared_db.ep_monitor_frame.refresh(None)
                    self.shared_db.plot_frame.refresh(None)
                    self.shared_db.log_frame.refresh(None)
                self.shared_db.last_refresh += 1
            self.shared_db.rl_sim_stop      = True
            self.__env_finish_check()
            self.shared_db.start_global_refresh()
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_simulation_run_cycle(self, *args):
        self.__first_training_run()
        if not self.shared_db.rl_sim_finished:
            self.shared_db.rl_sim_started   = True
            self.shared_db.rl_sim_stop      = False
            self.shared_db.rl_training.run_cycle() 
            self.shared_db.rl_sim_stop      = True
            self.__env_finish_check()
            self.shared_db.start_global_refresh()
    
    
  
    
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLPlot2D(SciUISubplot2D):
    """
    ...
    """

    C_NAME                      = 'RL Plot 2D'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        self.C_AX_RECTANGLE              = [0.10, 0.11, 0.85, 0.85]

        super().init_component()

        self.plot               = None
        self.min_val            = 1
        self.max_val            = 0
        self.val_range_changed  = False


## -------------------------------------------------------------------------------------------------
    def refresh_custom(self):
        super().refresh_custom()

        if self.plot == None:
            self.ax.grid()
            self.ax.plot([0],[0],color='blue', lw=1)
        self.ax.set_xlim(0, 100)
        self.ax.set_ylim(0, 1)
        
    
    
  
    
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLHeatmap2D(SciUISubplot2D):
    """
    ...
    """

    C_NAME                      = 'RL Heatmap 2D'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        self.plot               = None
        self.min_val            = 1
        self.max_val            = 0
        self.val_range_changed  = False
        
        self.random_data        = np.random.random((16, 16))
        self.title_heatmap      = None


## -------------------------------------------------------------------------------------------------
    def refresh_custom(self):
        super().refresh_custom()

        if self.plot == None:
            im = self.ax.imshow(self.random_data, cmap='winter', interpolation='nearest')
        
        
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLStatesToggle(SciUIFrameParam):
    """
    ...
    """

    C_NAME                  = 'States Toggle'
    C_AGENT_LABELS          = ['Agent 01','Agent 02','Agent 03','Agent 04','Agent 05']
    C_STATES_LABELS         = ['State 01','State 02']


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh  = True
        
        self.agent_var      = StringVar()
        self.states_var     = StringVar()


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        
        self.agent_box      = Combobox(self.frame, textvariable=self.agent_var, width=15, justify='left', state='normal', background='white', font=self.fontstyle, values=self.C_AGENT_LABELS)
        self.agent_box.grid(row=0, column=0, padx=5, pady=5, sticky='NW')
        self.agent_var.trace('w', self.__cb_agent_changed)
        
        self.state_box      = Combobox(self.frame, textvariable=self.states_var, width=15, justify='left', state='normal', background='white', font=self.fontstyle, values=self.C_STATES_LABELS)
        self.state_box.grid(row=0, column=1, padx=5, pady=5, sticky='NW')
        self.states_var.trace('w', self.__cb_states_changed)
  
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_agent_changed(self, *args):
        pass 
        
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_states_changed(self, *args):
        pass 
        
        
        
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLStatesView(SciUIFrameParam):
    """
    ...
    """

    C_NAME                          = 'States'
    C_AGENT_LABELS                  = ['Agent 01','Agent 02','Agent 03','Agent 04','Agent 05']
    C_STATES_LABELS                 = ['State 01','State 02']


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh              = True
        
        self.states_plt                 = RLPlot2D(self.shared_db, p_row=1, p_col=0, p_width_perc=0.36, p_height_perc=0.26, p_logging=self._level)
        self.states_plt.frame_visible   = False
        self.states_toggle              = RLStatesToggle(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5,
                                                     p_logging=self._level, p_height_perc=1)
        
        self.add_component(self.states_plt)
        self.add_component(self.states_toggle)


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        
        
        
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLActionsView(SciUIFrameParam):
    """
    ...
    """

    C_NAME                                  = 'Actions'
    C_AGENT_LABELS                          = ['Agent 01','Agent 02','Agent 03','Agent 04','Agent 05']


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh                  = True
        
        self.actions_plt                    = RLPlot2D(self.shared_db, p_row=1, p_col=0, p_width_perc=0.36, p_height_perc=0.26, p_logging=self._level)
        self.actions_plt.frame_visible      = False
        self.add_component(self.actions_plt)
        
        self.agent_var                      = StringVar()


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        
        self.agent_box                      = Combobox(self.frame, textvariable=self.agent_var, width=15, justify='left', state='normal', background='white', font=self.fontstyle, values=self.C_AGENT_LABELS)
        self.agent_box.grid(row=0, column=0, padx=15, pady=10, sticky='NW')
        self.agent_var.trace('w', self.__cb_agent_changed)
  
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_agent_changed(self, *args):
        pass 
        
        
        
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLRewardsView(SciUIFrameParam):
    """
    ...
    """

    C_NAME                              = 'Rewards'
    C_AGENT_LABELS                      = ['Agent 01','Agent 02','Agent 03','Agent 04','Agent 05','Total Reward']


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh              = True
        
        self.rewards_plt                = RLPlot2D(self.shared_db, p_row=1, p_col=0, p_width_perc=0.36, p_height_perc=0.26, p_logging=self._level)
        self.rewards_plt.frame_visible  = False
        self.add_component(self.rewards_plt)
        
        self.agent_var                  = StringVar()


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        
        self.agent_box                  = Combobox(self.frame, textvariable=self.agent_var, width=15, justify='left', state='normal', background='white', font=self.fontstyle, values=self.C_AGENT_LABELS)
        self.agent_box.grid(row=0, column=0, padx=15, pady=10, sticky='NW')
        self.agent_var.trace('w', self.__cb_agent_changed)
  
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_agent_changed(self, *args):
        pass 
        
        
        
        
        
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLGoalsCompletedView(SciUIFrameParam):
    """
    ...
    """

    C_NAME                              = 'Goals Completed'
    C_AGENT_LABELS                      = []


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.C_AGENT_LABELS.append(self.shared_db.rl_env)
        self.first_refresh              = True
        
        self.exp_rate_plt               = RLPlot2D(self.shared_db, p_row=1, p_col=0, p_width_perc=0.36, p_height_perc=0.26, p_logging=self._level)
        self.exp_rate_plt.frame_visible = False
        self.add_component(self.exp_rate_plt)
        
        self.agent_var                  = StringVar()


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        
        self.agent_box                  = Combobox(self.frame, textvariable=self.agent_var, width=15, justify='left', state='normal', background='white', font=self.fontstyle, values=self.C_AGENT_LABELS)
        self.agent_box.grid(row=0, column=0, padx=15, pady=10, sticky='NW')
        self.agent_var.trace('w', self.__cb_agent_changed)
  
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_agent_changed(self, *args):
        pass 
        
        
        
        
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLCustomView(SciUIFrameParam):
    """
    ...
    """

    C_NAME                              = 'Customs'
    C_AGENT_LABELS                      = []


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh              = True
        
        self.custom_plt                 = RLPlot2D(self.shared_db, p_row=1, p_col=0, p_width_perc=0.36, p_height_perc=0.26, p_logging=self._level)
        self.custom_plt.frame_visible   = False
        self.add_component(self.custom_plt)
        
        self.agent_var                  = StringVar()


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        
        self.agent_box                  = Combobox(self.frame, textvariable=self.agent_var, width=15, justify='left', state='normal', background='white', font=self.fontstyle, values=self.C_AGENT_LABELS)
        self.agent_box.grid(row=0, column=0, padx=15, pady=10, sticky='NW')
        self.agent_var.trace('w', self.__cb_agent_changed)
  
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_agent_changed(self, *args):
        pass 

        
        
        
        
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLHeatmapsView(SciUIFrameParam):
    """
    ...
    """

    C_NAME                                      = 'Performance Maps - Action and Reward'
    C_AGENT_LABELS                              = ['Agent 01','Agent 02','Agent 03','Agent 04','Agent 05']


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh                      = True
        
        self.maps_actions_plt                   = RLHeatmap2D(self.shared_db, p_row=1, p_col=0, p_width_perc=0.18, p_height_perc=0.26, p_logging=self._level)
        self.maps_actions_plt.title_heatmap     = "Performance Maps - Actions"
        self.maps_actions_plt.frame_visible     = False
        
        self.maps_rewards_plt                   = RLHeatmap2D(self.shared_db, p_row=1, p_col=1, p_width_perc=0.18, p_height_perc=0.26, p_logging=self._level)
        self.maps_rewards_plt.title_heatmap     = "Performance Maps - Rewards"
        self.maps_rewards_plt.frame_visible     = False
        
        self.add_component(self.maps_actions_plt)
        self.add_component(self.maps_rewards_plt)
        
        self.agent_var                          = StringVar()


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        
        self.agent_box                          = Combobox(self.frame, textvariable=self.agent_var, width=15, justify='left', state='normal', background='white', font=self.fontstyle, values=self.C_AGENT_LABELS)
        self.agent_box.grid(row=0, column=0, padx=15, pady=10, sticky='NW')
        self.agent_var.trace('w', self.__cb_agent_changed)
  
        
 ## -------------------------------------------------------------------------------------------------
    def __cb_agent_changed(self, *args):
        pass 
        




    
    
    
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLGeneralView(SciUIFrame):
    """
    ...
    """

    C_NAME          = 'General View'


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh  = True

        self.add_component(RLStatesView(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=8,
                                        p_padx=10, p_logging=self._level, p_visible=True))
        self.add_component(RLActionsView(p_shared_db=self.shared_db, p_row=0, p_col=1, p_pady=8,
                                         p_padx=10, p_logging=self._level, p_visible=True))
        self.add_component(RLRewardsView(p_shared_db=self.shared_db, p_row=1, p_col=0, p_pady=8,
                                         p_padx=10, p_logging=self._level, p_visible=True))
        self.add_component(RLGoalsCompletedView(p_shared_db=self.shared_db, p_row=1, p_col=1, p_pady=8,
                                                p_padx=10, p_logging=self._level, p_visible=True))


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        

        




    
    
    
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLEnvironmentView(SciUIFrame):
    """
    ...
    """

    C_NAME          = 'Environment View'


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh  = True

        self.add_component(RLCustomView(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=8,
                                        p_padx=10, p_logging=self._level, p_visible=True))


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        

        




    
    
    
    
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLAgentView(SciUIFrame):
    """
    ...
    """

    C_NAME          = 'RL Agent(s) View'


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh  = True

        self.add_component(RLHeatmapsView(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=8,
                                          p_padx=10, p_logging=self._level, p_visible=True))


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)

        



## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLView(SciUITabCTRL):
    """
    ...
    """

    C_NAME          = 'Real-Time Monitoring'
    
    C_FONT_FAMILY   = 'Lucida Grande'
    C_FONT_SIZE     = 10 


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh  = True
        self.fontstyle  = tkFont.Font(family=self.C_FONT_FAMILY, size=self.C_FONT_SIZE)
        self.tab_list   = []
        self.tab_ctrl   = None 
        self.add_tab('General view', RLGeneralView(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5, p_logging=self._level))
        self.add_tab('Environment view', RLEnvironmentView(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5, p_logging=self._level))
        self.add_tab('RL Agent(s) view', RLAgentView(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5, p_logging=self._level))

## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)
        self.tab_ctrl.config(width=1440, height=680)
        
        
        
        
        

## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLEpMonitoring(SciUIFrameParam):
    """
    ...
    """

    C_NAME          = 'Current Status'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh          = True
        self.ep_monitor_var         = IntVar()
        self.cycletime_monitor_var  = IntVar()
        self.run_monitor_var        = StringVar()
        self.ep_monitor_var.set(0)
        self.cycletime_monitor_var.set(0)
        self.run_monitor_var.set('Stop')


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        if self.first_refresh: 
            self.first_refresh  = False
        else: 
            try:
                self.ep_monitor_var.set(self.shared_db.rl_training._episode_id)
                self.cycletime_monitor_var.set(self.shared_db.rl_training._cycle_id)
            except:
                self.ep_monitor_var.set(0)
                self.cycletime_monitor_var.set(0)
                
            if self.shared_db.rl_sim_finished:
                self.run_monitor_var.set('Done')
                self.shared_db.bg_color_run_button = "yellow"
            elif self.shared_db.rl_sim_stop:
                self.run_monitor_var.set('Stop')
                self.shared_db.bg_color_run_button = "red"
            else:
                self.run_monitor_var.set('Running')
                self.shared_db.bg_color_run_button = "green"
                

        super().refresh(p_parent_frame=p_parent_frame)
        
        self.env_label_1        = Label(self.frame, text='Episode', pady=5, font=self.fontstyle, width=15, height=5, anchor='s').grid(row=0, column=0, sticky='s')
        self.alg_label_1        = Label(self.frame, text='Cycle Time', pady=5, font=self.fontstyle, width=15, height=5, anchor='s').grid(row=0, column=1, sticky='s')
        self.run_label_1        = Label(self.frame, text='Simulation (on/off)', pady=5, font=self.fontstyle, width=20, height=5, anchor='s').grid(row=0, column=2, sticky='s')
    
        self.env_label_2        = Label(self.frame, textvariable=self.ep_monitor_var, pady=4, font=self.fontstyle, width=15, height=5, anchor='n').grid(row=1, column=0, sticky='n')
        self.alg_label_2        = Label(self.frame, textvariable=self.cycletime_monitor_var, pady=4, font=self.fontstyle, width=15, height=5, anchor='n').grid(row=1, column=1, sticky='n')
        self.run_label_2        = Label(self.frame, textvariable=self.run_monitor_var, pady=4, font=self.fontstyle, width=20, height=1, anchor='n', borderwidth=2, relief='raised', bg=self.shared_db.bg_color_run_button).grid(row=1, column=2, sticky='n')
 

        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLLoggingSelection(SciUIFrameParam):
    """
    ...
    """

    C_NAME          = 'Logging settings'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh      = True
        self.set_env_log        = BooleanVar()
        self.set_agent_log      = BooleanVar()
        self.set_process_log    = BooleanVar()
        self.set_train_log      = BooleanVar()
        self.set_env_log.set(self.shared_db.rl_log_env)
        self.set_agent_log.set(self.shared_db.rl_log_agent)
        self.set_process_log.set(self.shared_db.rl_log_process)
        self.set_train_log.set(self.shared_db.rl_log_training)


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        if self.first_refresh: 
            self.first_refresh  = False
        else: 
            if self.shared_db.rl_log_state_box:
                self.shared_db.rl_log_state_btn = 'normal'
            else:
                self.shared_db.rl_log_state_btn = 'disabled'

        super().refresh(p_parent_frame=p_parent_frame)
        
        self.set_env_button     = Checkbutton(self.frame, text = "Environment", justify='left', variable=self.set_env_log, height=2, width=12, anchor='w', state=self.shared_db.rl_log_state_btn).grid(row=0, column=0, padx=5, pady=0, sticky='nw')
        self.set_env_log.trace('w', self.__cb_set_env_changed)
        
        self.set_agent_button   = Checkbutton(self.frame, text = "Agent(s)", justify='left', variable=self.set_agent_log, height=2, width=12, anchor='w', state=self.shared_db.rl_log_state_btn).grid(row=1, column=0, padx=5, pady=0, sticky='nw')
        self.set_agent_log.trace('w', self.__cb_set_agent_changed)
        
        self.set_process_button = Checkbutton(self.frame, text = "Process", justify='left', variable=self.set_process_log, height=2, width=12, anchor='w', state=self.shared_db.rl_log_state_btn).grid(row=2, column=0, padx=5, pady=0, sticky='nw')
        self.set_process_log.trace('w', self.__cb_set_process_changed)
        
        self.set_train_button   = Checkbutton(self.frame, text = "Training", justify='left', variable=self.set_train_log, height=2, width=12, anchor='w', state=self.shared_db.rl_log_state_btn).grid(row=3, column=0, padx=5, pady=0, sticky='nw')
        self.set_train_log.trace('w', self.__cb_set_train_changed)
        
        Label(self.frame, text = " ", justify='left', height=1, width=12, anchor='w').grid(row=4, column=0, padx=5, pady=2.4, sticky='nw')
        
## -------------------------------------------------------------------------------------------------
    def __cb_set_env_changed(self, *args):
        try:
            val = self.set_env_log.get()
        except:
            return
        
        self.shared_db.rl_log_env = val
        self.set_env_log.set(self.shared_db.rl_log_env)
        self.shared_db.rl_scenario._env.switch_logging(self.set_env_log.get())
        
## -------------------------------------------------------------------------------------------------
    def __cb_set_agent_changed(self, *args):
        try:
            val = self.set_agent_log.get()
        except:
            return
        
        self.shared_db.rl_log_agent = val
        self.set_agent_log.set(self.shared_db.rl_log_agent)
        self.shared_db.rl_scenario._agent.switch_logging(self.set_agent_log.get())
        
## -------------------------------------------------------------------------------------------------
    def __cb_set_process_changed(self, *args):
        try:
            val = self.set_process_log.get()
        except:
            return
        
        self.shared_db.rl_log_process = val
        self.set_process_log.set(self.shared_db.rl_log_process)
        self.shared_db.rl_scenario.switch_logging(self.set_process_log.get())
        
## -------------------------------------------------------------------------------------------------
    def __cb_set_train_changed(self, *args):
        try:
            val = self.set_train_log.get()
        except:
            return
        
        self.shared_db.rl_log_training = val
        self.set_train_log.set(self.shared_db.rl_log_training)
        self.shared_db.rl_training.switch_logging(self.set_train_log.get())
        
        
        

## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLLogging(SciUIFrameParam):
    """
    ...
    """

    C_NAME          = 'Simulation Log'
    C_NUM_ROWS      = 10

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.first_refresh  = True


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        if self.first_refresh: 
            self.first_refresh = False
        else: 
            return

        super().refresh(p_parent_frame=p_parent_frame)
        
        self.log_frame = ScrolledText(self.frame, height=9, width=177, wrap=WORD)
        
        ### 1. Header ###
        self.log_frame.config(pady=20, padx=20, font=self.fontstyle, state='normal')
        self.log_frame.insert(INSERT, "Logging taken from MLPro class!\n")
        # self.log_frame.insert(END, "Logging taken from MLPro class!\n")
        # self.log_frame.yview(END)
        self.log_frame.config(state='disabled')
        self.log_frame.pack()
        self.log_frame.insert(INSERT, "Logging taken from MLPro class!\n")
        self.log_frame.pack()
        
        
        
        
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLFrameTopLeft(SciUIFrame):
    """
    ...
    """

    C_NAME                      = 'RL Frame Top-left'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        self.add_component(RLDetails(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5,
                                     p_logging=self._level, p_visible=True, p_height_perc=1))

        self.add_component(RLParam(p_shared_db=self.shared_db, p_row=1, p_col=0, p_pady=5,
                                     p_logging=self._level, p_visible=True, p_height_perc=1))
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLFrameTopRight(SciUIFrame):
    """
    ...
    """

    C_NAME                      = 'RL Frame Top-Right'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        
        self.shared_db.plot_frame = RLView(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5,
                                           p_logging=self._level, p_visible=True, p_height_perc=1)
        self.add_component(self.shared_db.plot_frame)

        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLFrameTop(SciUIFrame):
    """
    ...
    """

    C_NAME                      = 'RL Frame Top'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        self.add_component(RLFrameTopLeft(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5,
                                          p_logging=self._level))

        self.add_component(RLFrameTopRight(p_shared_db=self.shared_db, p_row=0, p_col=1, p_pady=5,
                                           p_logging=self._level))
        
        
        
        
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLFrameBottomLeft(SciUIFrame):
    """
    ...
    """

    C_NAME                      = 'RL Frame Bottom-left'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        
        self.shared_db.ep_monitor_frame = RLEpMonitoring(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5,p_padx=5,
                                                         p_logging=self._level, p_visible=True, p_height_perc=1)
        self.add_component(self.shared_db.ep_monitor_frame)
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLFrameBottomRight(SciUIFrame):
    """
    ...
    """

    C_NAME                      = 'RL Frame Bottom-Right'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        
        self.add_component(RLLoggingSelection(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5,p_padx=5,
                                              p_logging=self._level, p_visible=True, p_height_perc=1))
        self.shared_db.log_frame = RLLogging(p_shared_db=self.shared_db, p_row=0, p_col=1, p_pady=5,p_padx=5,
                                             p_logging=self._level, p_visible=True, p_height_perc=1) 
        self.add_component(self.shared_db.log_frame)


        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLFrameBottom(SciUIFrame):
    """
    ...
    """

    C_NAME                      = 'RL Frame Bottom'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        self.add_component(RLFrameBottomLeft(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5,
                                             p_logging=self._level))
        self.add_component(RLFrameBottomRight(p_shared_db=self.shared_db, p_row=0, p_col=1, p_pady=5,
                                              p_logging=self._level))


        
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLInteractiveUI(SciUIFrame): 
    """
    Main class for the RL interactive UI.
    """

    C_NAME                      = 'RL Interactive UI'
    C_ENV_LABELS                = ''
    C_ALG_LABELS                = ''
    
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, p_shared_db:SciUISharedDB, p_row, p_col, p_title=None, p_width_perc=0.0,
                 p_height_perc=0.0, p_visible=False, p_padx=5, p_pady=0, p_sticky='NW', p_logging=True,
                 p_refresh_rate=100): 
        """
        Parameters:
            p_shared_db         Shared DB object for scenario-internal communication
            p_row               Row in superior frame
            p_col               Column in superior frame
            p_title             Frame title
            p_width_perc        Frame width in percent
            p_height_perc       Frame height in percent
            p_visible           Boolean switch for visibility
            p_padx              Horizontal distance to neighbor frames
            p_pady              Vertical distance to neighbor frames
            p_sticky            Tkinter sticky parameter 
            p_logging           Boolean switch for logging
            p_refresh_rate      Refresh rate per cycle time
        """

        self.row                        = p_row
        self.col                        = p_col
        self.width_perc                 = p_width_perc              # Frame width in percent of main window width
        self.height_perc                = p_height_perc             # Frame height in percent of main window height
        self.width_pix                  = p_shared_db.window_width * self.width_perc
        self.height_pix                 = p_shared_db.window_height * self.height_perc
        self.frame_visible              = p_visible

        if p_title == None:
            self.frame_text             = self.C_NAME
        else:
            self.frame_text             = p_title

        self.frame                      = None
        self.components                 = []
        self.sticky                     = p_sticky
        self.padx                       = p_padx
        self.pady                       = p_pady
        self.popup_menu                 = None
        self.refresh_rate               = p_refresh_rate
        
        super().__init__(p_shared_db, p_row, p_col, p_logging=p_logging) 
        
        
## -------------------------------------------------------------------------------------------------
    @staticmethod
    def enrich_shared_db(p_shared_db:SciUISharedDB):
        p_shared_db.rlui_selected_scenario  = False
        p_shared_db.refresh_rate            = 0
        p_shared_db.last_refresh            = 0
        
        p_shared_db.rl_env                  = ""
        p_shared_db.rl_learning_alg         = ""
        p_shared_db.rl_scenario_class       = None
        p_shared_db.rl_scenario             = None
        p_shared_db.rl_training             = None
        
        p_shared_db.rl_ep_limit             = 0
        p_shared_db.rl_cycle_limit          = 0
        p_shared_db.rl_collect_states       = True
        p_shared_db.rl_collect_rewards      = True
        p_shared_db.rl_collect_actions      = True
        p_shared_db.rl_collect_training     = True
        p_shared_db.rl_save_data            = True
        
        p_shared_db.rl_log_env              = False
        p_shared_db.rl_log_agent            = False
        p_shared_db.rl_log_process          = False
        p_shared_db.rl_log_training         = True
        
        p_shared_db.rl_sim_started          = False
        p_shared_db.rl_sim_stop             = True
        p_shared_db.rl_sim_finished         = False
        p_shared_db.rl_refresh_rate         = 10
        
        p_shared_db.rl_sce_state_btn        = True
        p_shared_db.rl_param_state_box      = 'normal'
        p_shared_db.rl_run_state_btn        = 'normal'
        p_shared_db.rl_log_state_box        = True
        p_shared_db.rl_log_state_btn        = 'normal'
        p_shared_db.bg_color_run_button     = 'red'
        
        p_shared_db.log_frame               = None
        p_shared_db.plot_frame              = None
        p_shared_db.ep_monitor_frame        = None


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.add_component(RLFrameTop(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady=5,
                                      p_logging=self._level))
        self.add_component(RLFrameBottom(p_shared_db=self.shared_db, p_row=1, p_col=0, p_pady=5,
                                         p_logging=self._level))

 ## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        self.shared_db.refresh_rate                = self.refresh_rate
        if self.shared_db.rl_sim_started == False and self.shared_db.rl_sim_stop == True:
            self.shared_db.rl_sce_state_btn        = True
            self.shared_db.rl_param_state_box      = 'normal'
            self.shared_db.rl_run_state_btn        = 'normal'
            self.shared_db.rl_log_state_box        = True
        elif self.shared_db.rl_sim_started == True and self.shared_db.rl_sim_stop == False:
            self.shared_db.rl_sce_state_btn        = False
            self.shared_db.rl_param_state_box      = 'disabled'
            self.shared_db.rl_run_state_btn        = 'disabled'
            self.shared_db.rl_log_state_box        = False
        elif self.shared_db.rl_sim_started == True and self.shared_db.rl_sim_stop == True:
            self.shared_db.rl_sce_state_btn        = False
            self.shared_db.rl_param_state_box      = 'disabled'
            self.shared_db.rl_run_state_btn        = 'normal'
            self.shared_db.rl_log_state_box        = True
            if self.shared_db.rl_sim_finished:
                self.shared_db.rl_run_state_btn    = 'disabled'
                self.shared_db.rl_log_state_box    = False

        super().refresh(p_parent_frame=p_parent_frame)
    
    
    
    
    


import rospy
import gym
from gym.utils import seeding
from .controllers_connection import ControllersConnection
from geometry_msgs.msg import Pose
from tf2_msgs.msg import TFMessage
from openai_ros.msg import RLExperimentInfo

class RobotRealEnv(gym.Env):

    def __init__(self, robot_name_space, controllers_list, reset_controls) -> None:
        # To reset Simulations
        rospy.logdebug("START init RobotRealEnv")
        self.controllers_object = ControllersConnection(namespace=robot_name_space, controllers_list=controllers_list)
        self.reset_controls = reset_controls
        self.seed()

        self.episode_num = 0
        self.cumulated_episode_reward = 0
        rospy.Subscriber("/tf", TFMessage, self._tf_real_sub)
        self._reward_pub = rospy.Publisher('/openai/reward', RLExperimentInfo, queue_size=1)
        self._tcp_real_pose_pub = rospy.Publisher('/tcp_pose_real', Pose, queue_size=10)
        self._tcp_real_msg = Pose()

        if self.reset_controls:
           self.controllers_object.reset_controllers()
        rospy.logdebug("END init RobotRealEnv")

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action):
        """
        Function executed each time step.
        Here we get the action execute it in a time step and retrieve the
        observations generated by that action.
        :param action:
        :return: obs, reward, done, info
        """

        """
        Here we should convert the action num to movement action, execute the action in the
        simulation and get the observations result of performing that action.
        """
        rospy.logdebug("START STEP OpenAIROS")

        self._set_action(action)

        obs = self._get_obs()
        done = self._is_done(obs)
        info = {}
        reward = self._compute_reward(obs, done)
        self.cumulated_episode_reward += reward

        rospy.logdebug("END STEP OpenAIROS")

        return obs, reward, done, info

    def reset(self):
        rospy.logdebug("Reseting RobotRealEnvironment")
        self._reset_env()
        self._init_env_variables()
        self._update_episode()
        obs = self._get_obs()
        rospy.logdebug("END Reseting RobotRealEnvironment")
        return obs

    def close(self):
        """
        Function executed when closing the environment.
        Use it for closing GUIS and other systems that need closing.
        :return:
        """
        rospy.logdebug('Closing RobotRealEnvironment')
        rospy.signal_shutdown('Closing RobotRealEnvironment')

    def _update_episode(self):
        """
        Publishes the cumulated reward of the episode and
        increases the episode number by one.
        :return:
        """
        rospy.logwarn("PUBLISHING REWARD...")
        self._publish_reward_topic(
                                    self.cumulated_episode_reward,
                                    self.episode_num
                                    )
        rospy.logwarn("PUBLISHING REWARD...DONE="+str(self.cumulated_episode_reward)+",EP="+str(self.episode_num))

        self.episode_num += 1
        self.cumulated_episode_reward = 0

    def _tf_real_sub(self, data):
        if data.transforms[0].child_frame_id == "tool0_controller":
            self._tcp_real_msg.position.x = data.transforms[0].transform.translation.x
            self._tcp_real_msg.position.y = data.transforms[0].transform.translation.y
            self._tcp_real_msg.position.z = data.transforms[0].transform.translation.z
            self._tcp_real_msg.orientation = data.transforms[0].transform.rotation

    def _publish_reward_topic(self, reward, episode_number=1):
        """
        This function publishes the given reward in the reward topic for
        easy access from ROS infrastructure.
        :param reward:
        :param episode_number:
        :return:
        """
        reward_msg = RLExperimentInfo()
        reward_msg.episode_number = episode_number
        reward_msg.episode_reward = reward
        self._reward_pub.publish(reward_msg)

    def _reset_env(self):
        """Resets a simulation
        """
        rospy.logdebug("RESET ENV START")
        rospy.logwarn("DONT RESET CONTROLLERS")
        self._check_all_systems_ready()
        self._set_init_pose()
        self._check_all_systems_ready()

        rospy.logdebug("RESET ENV END")
        return True

    def _set_init_pose(self):
        """Sets the Robot in its init pose
        """
        raise NotImplementedError()

    def _check_all_systems_ready(self):
        """
        Checks that all the sensors, publishers and other simulation systems are
        operational.
        """
        raise NotImplementedError()

    def _get_obs(self):
        """Returns the observation.
        """
        raise NotImplementedError()

    def _init_env_variables(self):
        """Inits variables needed to be initialised each time we reset at the start
        of an episode.
        """
        raise NotImplementedError()

    def _set_action(self, action):
        """Applies the given action to the simulation.
        """
        raise NotImplementedError()

    def _is_done(self, observations):
        """Indicates whether or not the episode is done ( the robot has fallen for example).
        """
        raise NotImplementedError()

    def _compute_reward(self, observations, done):
        """Calculates the reward to give based on the observations given.
        """
        raise NotImplementedError()

    def _env_setup(self, initial_qpos):
        """Initial configuration of the environment. Can be used to configure initial state
        and extract information from the simulation.
        """
        raise NotImplementedError()


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_gt_001_run_multi_player_with_own_policy_in_multicartpole_game_board.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-06-06  0.0.0     DA       Creation
## -- 2021-06-06  1.0.0     DA       Release of first version
## -- 2021-08-28  1.0.1     DA       Adjustments after changings on rl models
## -- 2021-09-11  1.0.1     MRD      Change Header information to match our new library name
## -- 2021-10-06  1.0.2     DA       Adjustments after changings on rl models
## -- 2021-11-15  1.1.0     DA       Refactoring 
## -- 2022-02-25  1.1.1     SY       Refactoring due to auto generated ID in class Dimension
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.1 (2022-02-25)

This module shows how to run an own multi-player with the enhanced multi-action game board 
MultiCartPole based on the OpenAI Gym CartPole environment.
"""


from mlpro.rl.models import *
from mlpro.gt.models import *
from mlpro.gt.pool.boards.multicartpole import MultiCartPolePGT
import random
import numpy as np





# 1 Implement your own agent policy
class MyPolicy (Policy):

    C_NAME      = 'MyPolicy'

    def set_random_seed(self, p_seed=None):
        random.seed(p_seed)


    def compute_action(self, p_state: State) -> Action:
        # 1 Create a numpy array for your action values 
        my_action_values = np.zeros(self._action_space.get_num_dim())

        # 2 Computing action values is up to you...
        for d in range(self._action_space.get_num_dim()):
            my_action_values[d] = random.random() 

        # 3 Return an action object with your values
        return Action(self._id, self._action_space, my_action_values)


    def _adapt(self, *p_args) -> bool:
        # 1 Adapting the internal policy is up to you...
        self.log(self.C_LOG_TYPE_W, 'Sorry, I am a stupid agent...')

        # 2 Only return True if something has been adapted...
        return False




# 2 Implement your own game
class MyGame (Game):

    C_NAME      = 'Matrix'

    def _setup(self, p_mode, p_ada, p_logging):

        # 1 Setup Multi-Player Environment (consisting of 3 OpenAI Gym Cartpole envs)
        self._env   = MultiCartPolePGT(p_num_envs=3, p_logging=p_logging)


        # 2 Setup Multi-Player

        # 2.1 Create empty Multi-Player
        multi_player = MultiPlayer(
            p_name='Human Beings',
            p_ada=True,
            p_logging=p_logging
        )

        # 2.2 Add Single-Player #1 with own policy (controlling sub-environment #1)
        ss_ids = self._env.get_state_space().get_dim_ids()
        as_ids = self._env.get_action_space().get_dim_ids()
        multi_player.add_player(
            p_player=Player(
                p_policy=MyPolicy(
                    p_observation_space=self._env.get_state_space().spawn([ss_ids[0],ss_ids[1],ss_ids[2],ss_ids[3]]),
                    p_action_space=self._env.get_action_space().spawn([as_ids[0]]),
                    p_ada=True,
                    p_logging=p_logging
                ),
                p_name='Neo',
                p_id=0,
                p_ada=True,
                p_logging=p_logging
            ),
            p_weight=0.3
        )


        # 2.3 Add Single-Player #2 with own policy (controlling sub-environments #2,#3)
        multi_player.add_player(
            p_player=Player(
                p_policy=MyPolicy(
                    p_observation_space=self._env.get_state_space().spawn([ss_ids[4],ss_ids[5],ss_ids[6],ss_ids[7],ss_ids[8],ss_ids[9],ss_ids[10],ss_ids[11]]),
                    p_action_space=self._env.get_action_space().spawn([as_ids[1],as_ids[2]]),
                    p_ada=True,
                    p_logging=p_logging
                ),
                p_name='Trinity',
                p_id=1,
                p_ada=True,
                p_logging=p_logging
            ),
            p_weight=0.7
        )

        # 2.4 Adaptive ML model (here: our multi-player) is returned
        return multi_player




# 3 Create game and run some cycles

if __name__ == "__main__":
    # 3.1 Parameters for demo mode
    logging     = Log.C_LOG_ALL
    visualize   = True
  
else:
    # 3.2 Parameters for internal unit test
    logging     = Log.C_LOG_NOTHING
    visualize   = False


mygame  = MyGame(
    p_mode=Mode.C_MODE_SIM,
    p_ada=True,
    p_cycle_limit=200,
    p_visualize=visualize,
    p_logging=logging )

mygame.run()


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.wrappers
## -- Module  : sb3.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-10-07  0.0.0     MRD      Creation
## -- 2021-10-07  1.0.0     MRD      Release of first version
## -- 2021-10-18  1.0.1     DA       Refactoring class WrPolicySB32MLPro
## -- 2021-10-18  1.1.0     MRD      SB3 Off Policy Wrapper on WrPolicySB32MLPro
## -- 2021-10-27  1.1.1     MRD      Mismatch datatype last_done on WrPolicySB32MLPro
## -- 2021-11-18  1.1.2     MRD      Put DummyEnv class outside the WrPolicySB32MLPro
## -- 2021-12-20  1.1.3     DA       Replaced calls get_done() by get_success()
## -- 2021-12-21  1.1.4     MRD      Refactor due to new update, regarding initial and terminal
## --                                state 
## -- 2022-01-18  1.1.5     MRD      Fix mismatch Off Policy Algorithm by adding more additional
## --                                information on adapt_off_policy()
## -- 2022-01-20  1.1.6     MRD      Fix the bug due to new version of SB3 1.4.0
## -- 2022-02-25  1.1.7     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-05-31  1.1.8     SY       Enable the possibility to process reward type C_TYPE_EVERY_AGENT
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.8 (2022-05-31)
This module provides wrapper classes for reinforcement learning tasks.
"""

import gym
import torch
from stable_baselines3.common import utils
from stable_baselines3.common.on_policy_algorithm import OnPolicyAlgorithm
from mlpro.rl.models import *


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class DummyEnv(gym.Env):
    """
    Dummy class for Environment. This is required due to some of the SB3 Policy Algorithm requires to have
    an Environment. As for now, it only needs the observation space and the action space.
    """

    def __init__(self, p_observation_space, p_action_space) -> None:
        super().__init__()
        self.observation_space = p_observation_space
        self.action_space = p_action_space


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class WrPolicySB32MLPro(Policy):
    """
    This class provides a policy wrapper from Standard Baselines 3 (SB3).
    Especially On-Policy Algorithm
    """

    C_TYPE = 'SB3 Policy'

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_sb3_policy, p_cycle_limit, p_observation_space, p_action_space, p_ada=True, p_logging=True):
        """
        Args:
            p_sb3_policy : SB3 Policy
            p_observation_space : Observation Space
            p_action_space : Environment Action Space
            p_ada (bool, optional): Adaptability. Defaults to True.
            p_logging (bool, optional): Logging. Defaults to True.
        """

        super().__init__(p_observation_space, p_action_space, p_ada=p_ada, p_logging=p_logging)

        self.sb3 = p_sb3_policy
        self.last_buffer_element = None
        self.last_done = False

        # Variable preparation for SB3
        action_space = None
        observation_space = None

        # Check if action is Discrete or Box
        action_dim = self.get_action_space().get_num_dim()
        id_dim = self.get_action_space().get_dim_ids()[0]
        if len(self.get_action_space().get_dim(id_dim).get_boundaries()) == 1:
            action_space = gym.spaces.Discrete(self.get_action_space().get_dim(id_dim).get_boundaries()[0])
        else:
            self.lows = []
            self.highs = []
            for dimension in range(action_dim):
                id_dim = self.get_action_space().get_dim_ids()[dimension]
                self.lows.append(self.get_action_space().get_dim(id_dim).get_boundaries()[0])
                self.highs.append(self.get_action_space().get_dim(id_dim).get_boundaries()[1])

            action_space = gym.spaces.Box(
                low=np.array(self.lows, dtype=np.float32),
                high=np.array(self.highs, dtype=np.float32),
                shape=(action_dim,),
                dtype=np.float32
            )

        # Check if state is Discrete or Box
        observation_dim = self.get_observation_space().get_num_dim()
        id_dim = self.get_observation_space().get_dim_ids()[0]
        if len(self.get_observation_space().get_dim(id_dim).get_boundaries()) == 1:
            observation_space = gym.spaces.Discrete(self.get_observation_space().get_dim(id_dim).get_boundaries()[0])
        else:
            lows = []
            highs = []
            for dimension in range(observation_dim):
                id_dim = self.get_observation_space().get_dim_ids()[dimension]
                lows.append(self.get_observation_space().get_dim(id_dim).get_boundaries()[0])
                highs.append(self.get_observation_space().get_dim(id_dim).get_boundaries()[1])

            observation_space = gym.spaces.Box(
                low=np.array(lows, dtype=np.float32),
                high=np.array(highs, dtype=np.float32),
                shape=(observation_dim,),
                dtype=np.float32
            )

        # Create Dummy Env
        self.sb3.env = DummyEnv(observation_space, action_space)

        # Setup SB3 Model
        self.sb3.observation_space = observation_space
        self.sb3.action_space = action_space
        self.sb3.n_envs = 1

        if isinstance(p_sb3_policy, OnPolicyAlgorithm):
            self.compute_action = self._compute_action_on_policy
            self._add_buffer = self._add_buffer_on_policy
            self._adapt = self._adapt_on_policy
            self.clear_buffer = self._clear_buffer_on_policy
            self._buffer = self.sb3.rollout_buffer
        else:
            self.compute_action = self._compute_action_off_policy
            self._add_buffer = self._add_buffer_off_policy
            self._adapt = self._adapt_off_policy
            self.clear_buffer = self._clear_buffer_off_policy
            self._buffer = self.sb3.replay_buffer
            self.collected_steps = 0

        self.sb3._setup_model()
        self.sb3._total_timesteps = p_cycle_limit
        self.sb3._logger = utils.configure_logger()

    def _compute_action_on_policy(self, p_obs: State) -> Action:
        obs = p_obs.get_values()

        if self._adaptivity:
            if not isinstance(obs, torch.Tensor):
                if isinstance(obs, list):
                    obs = torch.Tensor(obs).reshape(1, len(obs)).to(self.sb3.device)
                else:
                    obs = torch.Tensor(obs).reshape(1, obs.size).to(self.sb3.device)

            with torch.no_grad():
                actions, values, log_probs = self.sb3.policy.forward(obs)

            actions = actions.cpu().numpy()

            clipped_actions = actions
            # Clip the actions to avoid out of bound error
            if isinstance(self.sb3.action_space, gym.spaces.Box):
                clipped_actions = np.clip(actions, self.lows, self.highs)

            # Action Step
            action = clipped_actions.flatten()
            action = Action(self._id, self._action_space, action)

            # Action Buffer
            action_buffer = actions.flatten()
            action_buffer = Action(self._id, self._action_space, action_buffer)

            # Add to additional_buffer_element
            self.additional_buffer_element = dict(action=action_buffer, value=values, action_log=log_probs)
        else:
            if isinstance(obs, list):
                obs = torch.Tensor(obs).reshape(1, len(obs)).to(self.sb3.device)
            else:
                obs = torch.Tensor(obs).reshape(1, obs.size).to(self.sb3.device)
            action, _ = self.sb3.predict(obs, deterministic=True)

            action = action.flatten()
            action = Action(self._id, self._action_space, action)
        return action

    def _compute_action_off_policy(self, p_obs: State) -> Action:
        self.sb3._last_obs = p_obs.get_values()
        action, buffer_action = self.sb3._sample_action(self.sb3.learning_starts)

        action = action.flatten()
        action = Action(self._id, self._action_space, action)

        buffer_action = buffer_action.flatten()
        buffer_action = Action(self._id, self._action_space, buffer_action)

        # Add to additional_buffer_element
        self.additional_buffer_element = dict(action=buffer_action)
        return action

    def _adapt_off_policy(self, *p_args) -> bool:
        # Add to buffer
        self._add_buffer(p_args[0])

        # Should Collect more steps
        if self.collected_steps < self.sb3.train_freq.frequency:
            return False

        # Check Step > Warp Up Phase
        if self.sb3.num_timesteps > 0 and self.sb3.num_timesteps > self.sb3.learning_starts:
            gradient_steps = self.sb3.gradient_steps if self.sb3.gradient_steps >= 0 else self.collected_steps
            if gradient_steps > 0:
                self.sb3.train(batch_size=self.sb3.batch_size, gradient_steps=gradient_steps)
                self.collected_steps = 0
                return True

        self.collected_steps = 0
        return False

    def _adapt_on_policy(self, *p_args) -> bool:
        # Add to buffer
        self._add_buffer(p_args[0])

        # Adapt only when Buffer is full
        if not self.sb3.rollout_buffer.full:
            self.log(self.C_LOG_TYPE_I, 'Buffer is not full yet, keep collecting data!')
            return False

        last_obs = torch.Tensor(np.array([self.last_buffer_element.get_data()["state_new"].get_values()])).to(
            self.sb3.device)
        last_done = self.last_buffer_element.get_data()["state_new"].get_terminal()
        last_done = np.array([last_done])

        # Get the next value from the last observation
        with torch.no_grad():
            last_values = self.sb3.policy.predict_values(last_obs)

        # Compute Return and Advantage
        self.sb3.rollout_buffer.compute_returns_and_advantage(last_values=last_values, dones=last_done)

        # Train
        self.sb3.train()

        # Clear Buffer After Update
        self.sb3.rollout_buffer.reset()

        return True

    def _clear_buffer_on_policy(self):
        self.sb3.rollout_buffer.reset()

    def _clear_buffer_off_policy(self):
        self.sb3.replay_buffer.reset()

    def _add_buffer_off_policy(self, p_buffer_element: SARSElement):
        """
        Redefine add_buffer function. Instead of adding to MLPro SARBuffer, we are using
        internal buffer from SB3 for off_policy.
        """

        # Add num_collected_steps
        self.collected_steps += 1
        self.sb3.num_timesteps += 1

        self.last_buffer_element = self._add_additional_buffer(p_buffer_element)
        datas = self.last_buffer_element.get_data()

        info = {}

        if datas["state_new"].get_terminal() and datas["state_new"].get_timeout():
            info["TimeLimit.truncated"] = True

        self.sb3.replay_buffer.add(
            datas["state"].get_values(),
            datas["state_new"].get_values(),
            datas["action"].get_sorted_values(),
            datas["reward"].get_overall_reward(),
            datas["state_new"].get_terminal(),
            [info])

        self.sb3._update_current_progress_remaining(self.sb3.num_timesteps, self.sb3._total_timesteps)
        self.sb3._on_step()

    def _add_buffer_on_policy(self, p_buffer_element: SARSElement):
        """
        Redefine add_buffer function. Instead of adding to MLPro SARBuffer, we are using
        internal buffer from SB3 for on_policy.
        """
        self.sb3.num_timesteps += 1
        self.last_buffer_element = self._add_additional_buffer(p_buffer_element)
        datas = self.last_buffer_element.get_data()

        try:        
            rewards = datas["reward"].get_overall_reward()
        except:
            rewards = datas["reward"].get_agent_reward(self._id)

        if datas["state_new"].get_terminal() and datas["state_new"].get_timeout():
            terminal_obs = torch.Tensor(np.array([datas["state_new"].get_values()])).to(self.sb3.device)
            with torch.no_grad():
                terminal_value = self.sb3.policy.predict_values(terminal_obs)[0]
            rewards += self.sb3.gamma * terminal_value.cpu()

        self.sb3.rollout_buffer.add(
            datas["state"].get_values(),
            datas["action"].get_sorted_values(),
            rewards,
            datas["state"].get_initial(),
            datas["value"],
            datas["action_log"])

    ## -------------------------------------------------------------------------------------------------
    def _add_additional_buffer(self, p_buffer_element: SARSElement):
        p_buffer_element.add_value_element(self.additional_buffer_element)
        return p_buffer_element



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.bf.ui.scui
## -- Module  : framework
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-04-13  0.0.0     DA       Creation
## -- 2021-07-05  1.0.0     DA       Release of first version
## -- 2021-07-29  1.1.0     DA       Class SciUITabCTRL added
## -- 2021-09-11  1.1.0     MRD      Change Header information to match our new library name
## -- 2022-01-06  1.1.1     DA       Adjustments for matplotlib 3.5
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.1 (2022-01-06)

SciUI framework classes to be reused in own SciUI scenarios. Needs Matplotlib 3.3 or higher.
"""


import os 
os.environ['KMP_DUPLICATE_LIB_OK']='True'       # ... to avoid dll-exception under Windows

from math import floor
from pathlib import Path


from tkinter import *
import tkinter.font as tkFont
from tkinter.filedialog import askdirectory
from tkinter import ttk
from tkinter.ttk import Combobox

import matplotlib
from matplotlib.figure import Figure
from matplotlib.widgets import Cursor
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg 
from mpl_toolkits.mplot3d.axes3d import Axes3D

from mlpro.bf.various import *





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUISharedDB(): 
    """
    Container for scenario-internal data exchange and communication. Can be extended while runtime 
    by consuming classes.
    """

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_root=None): 
        
        # 1 Root object
        self.root               = p_root

         # 2 Screen 
        self.screen_dpi         = 0

        # 3 Application window 
        self.window_height      = 0
        self.window_width       = 0 
        self.window_resized     = False
        self.window_fullscreen  = False
        
        # 4 Application control
        self.refresh_all        = IntVar()
        self.refresh_all.set(0)
        self.refresh_full       = True
        self.redraw             = True          # Switch for graphical output
        
        
## -------------------------------------------------------------------------------------------------
    def start_global_refresh(self):   
        self.refresh_all.set( 1 - self.refresh_all.get() )
        
        
             


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUIRoot(Log): 
    """
    SciUI root class with overarching properties.
    """
    
    pass





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUIWindow(SciUIRoot, Tk):
    """
    Root class for SciUI window apllications.
    """

    C_TYPE          = 'SciUI Window'
    C_NAME          = '????'

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_logging=False):
        SciUIRoot.__init__(self, p_logging=p_logging)
        Tk.__init__(self)
        self.title(self.C_NAME)


## -------------------------------------------------------------------------------------------------
    def start(self):
        self.mainloop()       





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUICursor (SciUIRoot, Cursor):
    """
    Enriched matplotlib cursor widget.
    """
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, ax, horizOn=True, vertOn=True, useblit=False, **lineprops):
        Cursor.__init__(self, ax, horizOn=horizOn, vertOn=vertOn, useblit=useblit, **lineprops)
        self.cb_button_press_event      = None
        self.cb_button_release_event    = None
        self.cb_motion_notify_event     = None
        self.st_button_press_event      = False
        self.st_button_release_event    = False
        self.st_motion_notify_event     = False

         
## -------------------------------------------------------------------------------------------------
    def connect_event(self, event, callback):
        if event == 'button_press_event':
            self.cb_button_press_event      = callback
            Cursor.connect_event(self, event, self.onbuttonpressed)
        elif event == 'button_release_event':
            self.cb_button_release_event    = callback
            Cursor.connect_event(self, event, self.onbuttonreleased)
        elif event == 'motion_notify_event':
            self.cb_motion_notify_event     = callback
            Cursor.connect_event(self, event, self.onmove)
        else:
            Cursor.connect_event(self, event, callback)           

        
## -------------------------------------------------------------------------------------------------
    def set_event_status(self, event, status):
        if event == 'button_press_event':
            self.st_button_press_event      = status
        elif event == 'button_release_event':
            self.st_button_release_event    = status
        elif event == 'motion_notify_event':
            self.st_motion_notify_event     = status        

        
## -------------------------------------------------------------------------------------------------
    def onmove(self, event):
        Cursor.onmove(self, event)
        if ( event.inaxes != self.ax ) or ( self.st_motion_notify_event == False ) or ( event.button == None ) or ( event.xdata == None ) or ( event.ydata == None ): return 
        if self.cb_motion_notify_event != None: self.cb_motion_notify_event(event)       
        
        
## -------------------------------------------------------------------------------------------------
    def onbuttonpressed(self, event):
        if ( event.inaxes != self.ax ) or ( self.st_button_press_event == False ) or ( event.dblclick ) or ( event.button != 1 ): return 
        self.cb_button_press_event(event)       

        
## -------------------------------------------------------------------------------------------------
    def onbuttonreleased(self, event):
        if ( event.inaxes != self.ax ) or ( self.st_button_release_event == False ): return 
        self.cb_button_release_event(event)       





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUITooltip(object):
    """
    Enriched tooltip class.
    """

## -------------------------------------------------------------------------------------------------
    def __init__(self, widget, text='widget info'):
        self.waittime = 500     #miliseconds
        self.wraplength = 180   #pixels
        self.widget = widget
        self.text = text
        self.widget.bind("<Enter>", self.enter)
        self.widget.bind("<Leave>", self.leave)
        self.widget.bind("<ButtonPress>", self.leave)
        self.id = None
        self.tw = None

    
## -------------------------------------------------------------------------------------------------
    def enter(self, event=None):
        self.schedule()


## -------------------------------------------------------------------------------------------------   
    def leave(self, event=None):
        self.unschedule()
        self.hidetip()


## -------------------------------------------------------------------------------------------------   
    def schedule(self):
        self.unschedule()
        self.id = self.widget.after(self.waittime, self.showtip)


## -------------------------------------------------------------------------------------------------   
    def unschedule(self):
        id = self.id
        self.id = None
        if id:
            self.widget.after_cancel(id)


## -------------------------------------------------------------------------------------------------   
    def showtip(self, event=None):
        x = y = 0
        x, y, cx, cy = self.widget.bbox("insert")
        x += self.widget.winfo_rootx() + 25
        y += self.widget.winfo_rooty() + 20
        # creates a toplevel window
        self.tw = Toplevel(self.widget)
        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)
        self.tw.wm_geometry("+%d+%d" % (x, y))
        label = Label(self.tw, text=self.text, justify='left',
                       background="#ffffff", relief='solid', borderwidth=1,
                       wraplength = self.wraplength)
        label.pack(ipadx=1)


## -------------------------------------------------------------------------------------------------   
    def hidetip(self):
        tw = self.tw
        self.tw= None
        if tw:
            tw.destroy()
            
            
            
            

## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUIComponent(SciUIRoot): 
    """
    Elementry screen object in SciUI framework.
    """

    C_TYPE          = 'SciUI Component'
    C_NAME          = ''
            
## -------------------------------------------------------------------------------------------------
    def __init__(self, p_shared_db:SciUISharedDB, p_logging=True):
        """
        Parameters:
            p_shared_db         Shared DB object for scenario-internal communication
            p_logging           Boolean switch for logging 
        """

        super().__init__(p_logging=p_logging)
        self.shared_db      = p_shared_db
        self.init_component()


## -------------------------------------------------------------------------------------------------
    def init_component(self): 
        """
        Initialization of component-specific elements at instance creation time. To be redefined.
        """

        pass

        
## -------------------------------------------------------------------------------------------------
    def get_name(self): 
        return self.C_NAME        
        
        
## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame=None): 
        """
        Refresh of all component-specific elements. To be redefined. Please call super().refresh()
        at the beginning of your own implementation.

        Parameters:
            p_parent_frame      Parent frame object
        """

        self.parent_frame = p_parent_frame
        self.log(self.C_LOG_TYPE_I, 'refresh')        





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUIFrame(SciUIComponent): 
    """
    Enriched wrapper class for the Tkinter (Label-)Frame class, based on the Tkinter grid positioning model.
    """
    
    C_TYPE          = 'SciUI Frame'
    C_NAME          = ''
        
## -------------------------------------------------------------------------------------------------
    def __init__(self, p_shared_db:SciUISharedDB, p_row, p_col, p_title=None, p_width_perc=0.0, p_height_perc=0.0, p_visible=False, p_padx=5, p_pady=0, p_sticky='NW', p_logging=True): 
        """
        Parameters:
            p_shared_db         Shared DB object for scenario-internal communication
            p_row               Row in superior frame
            p_col               Column in superior frame
            p_title             Frame title
            p_width_perc        Frame width in percent
            p_height_perc       Frame height in percent
            p_visible           Boolean switch for visibility
            p_padx              Horizontal distance to neighbor frames
            p_pady              Vertical distance to neighbor frames
            p_sticky            Tkinter sticky parameter 
            p_logging           Boolean switch for logging
        """

        self.row            = p_row
        self.col            = p_col
        self.width_perc     = p_width_perc              # Frame width in percent of main window width
        self.height_perc    = p_height_perc             # Frame height in percent of main window height
        self.width_pix      = p_shared_db.window_width * self.width_perc
        self.height_pix     = p_shared_db.window_height * self.height_perc
        self.frame_visible  = p_visible

        if p_title == None:
            self.frame_text     = self.C_NAME
        else:
            self.frame_text     = p_title

        self.frame          = None
        self.components     = []
        self.sticky         = p_sticky
        self.padx           = p_padx
        self.pady           = p_pady
        self.popup_menu     = None
        
        super().__init__(p_shared_db, p_logging=p_logging)       
        
        
## -------------------------------------------------------------------------------------------------
    def init_popup_menu(self): 
        """
        Initializes popup menu of the component. To be redefined.
        """
        
        pass

#        Sample popup menu coding
#         if self.popup_menu != None: return
        
#         self.popup_menu = Menu(self.shared_db.root, tearoff=False)
#         self.popup_menu.add_command(label='My first entry!') #, command=self.__cb_my_first_entry)           
        
        
## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.init_popup_menu()       
        
        
## -------------------------------------------------------------------------------------------------
    def cb_popup_menu(self, p_event):
        self.log(self.C_LOG_TYPE_I, 'popup menu opened')
        self.popup_menu.tk_popup(p_event.x_root, p_event.y_root)            
        
        
## -------------------------------------------------------------------------------------------------
    def determine_frame_size(self):
        self.width_pix      = 0
        self.height_pix     = 0
        
        if self.width_perc != 0.0:
            self.width_pix  = floor(self.shared_db.window_width * self.width_perc)

        if self.height_perc != 0.0:
            self.height_pix  = floor(self.shared_db.window_height * self.height_perc)
            
        if self.width_pix == 0:
            self.width_pix = self.height_pix

        if self.height_pix == 0:
            self.height_pix = self.width_pix
        
        
## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame=None):
        super().refresh(p_parent_frame)
        self.determine_frame_size()
        
        if self.frame == None:
            if self.frame_visible:
                self.frame = LabelFrame(self.parent_frame,text=self.frame_text, padx=self.padx, pady=self.pady) 
                self.frame.grid(row=self.row, column=self.col, sticky=self.sticky)               
            else:
                self.frame = Frame(self.parent_frame, padx=self.padx, pady=self.pady) 
                self.frame.grid(row=self.row, column=self.col, sticky=self.sticky)

            self.init_popup_menu()
            if self.popup_menu != None: 
                self.frame.bind("<Button-3>",self.cb_popup_menu)
            
        for comp in self.components: comp.refresh(self.frame)    
    
        
## -------------------------------------------------------------------------------------------------
    def add_component(self, p_component:SciUIComponent):
        self.components.append(p_component)             
        
        
## -------------------------------------------------------------------------------------------------
    def __del__(self):
        del self.components[:]
        
        if self.frame != None: 
            self.frame.grid_forget()
            self.frame.grid_remove()
            self.frame = None
              




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUITabCTRL(SciUIFrame):
    """
    Enriched wrapper class for the Tkinter tab control.
    """

    C_TYPE          = 'SciUI Tabs'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.tab_list   = []
        self.tab_ctrl   = None       


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        super().refresh(p_parent_frame=p_parent_frame)

        if self.tab_ctrl == None:
            self.tab_ctrl = ttk.Notebook(self.frame)
            self.tab_ctrl.grid(column=self.col, row=self.row)

            for tab in self.tab_list:
                tab[1] = ttk.Frame(self.tab_ctrl)
                self.tab_ctrl.add(tab[1], text=tab[0])

        for tab in self.tab_list:
            if tab[2] != None: tab[2].refresh(tab[1])


## -------------------------------------------------------------------------------------------------
    def add_component(self, p_component: SciUIComponent):
        pass

    
## -------------------------------------------------------------------------------------------------
    def add_tab(self, p_tab_name, p_component:SciUIComponent):
        self.tab_list.append([p_tab_name, None, p_component])





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUISubplotSaveDLG(SciUIWindow):
    """
    Small SciUI window application to choose folder and file name for saving a SciUI subplot.
    """

    C_NAME          = 'Save Plot'
    C_FONT_FAMILY   = 'Lucida Grande'
    C_FONT_SIZE     = 10 
    C_FILENAME      = 'myplot'
    C_SUFFIXES      = ['.pdf', '.png', '.svg']

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_fname=None, p_xpos=500, p_ypos=300, p_logging=False):
        """
        Parameters:
            p_fname         Default filename
            p_xpos          X position of window
            p_ypos          Y position of window
            p_logging       Boolean switch for logging
        """

        super().__init__(p_logging=p_logging)

        if ( p_fname is None ) or ( p_fname == '' ):
            self.filename   = self.C_FILENAME
        else:
            self.filename   = p_fname

        self.geometry("+%d+%d" %(p_xpos,p_ypos))
        self.attributes("-topmost", True)

        self.fontstyle  = tkFont.Font(family=self.C_FONT_FAMILY, size=self.C_FONT_SIZE)
        self.frame      = LabelFrame(self, padx=5, pady=5, font=self.fontstyle) 
        self.frame.grid(row=0, column=0)               

        # Row 1 - Destination folder
        Label(self.frame, text='Folder', font=self.fontstyle).grid(row=1, column=0, sticky='w')     
        self.folder_var = StringVar(self)
        self.folder_var.set(Path.home())
        self.folder     = Entry(self.frame, textvariable=self.folder_var, width=40, justify='left', state='readonly', font=self.fontstyle)
        self.folder.grid(row=1, column=1)
        self.but_chg_folder = Button(self.frame, text='Change', command=self.__cb_button_folder)
        self.but_chg_folder.grid(row=1, column=2)

        # Row 2 - Filename and suffix
        Label(self.frame, text='File name', font=self.fontstyle).grid(row=2, column=0, sticky='w')     
        self.fname_var  = StringVar(self)
        self.fname_var.set(self.filename)
        self.fname      = Entry(self.frame, textvariable=self.fname_var, width=40, justify='left', background='white', font=self.fontstyle)
        self.fname.grid(row=2, column=1)

        self.suffix_var = StringVar(self)
        self.suffix_var.set(self.C_SUFFIXES[0])
        self.suffix     = Combobox(self.frame, textvariable=self.suffix_var, width=4, justify='left', state='readonly', font=self.fontstyle, values=self.C_SUFFIXES)
        self.suffix.grid(row=2, column=2, sticky='w')

        # Row 3 - Empty
        Label(self.frame, text='', font=self.fontstyle).grid(row=3, column=0)     

        # Row 4 - Buttons Save/Cancel
        self.but_cancel = Button(self.frame, text='Cancel', width=6, command=self.__cb_button_cancel)
        self.but_cancel.grid(row=4, column=2)

        self.but_ok     = Button(self.frame, text='Save', width=6, command=self.__cb_button_ok)
        self.but_ok.grid(row=4, column=3)


## -------------------------------------------------------------------------------------------------
    def get_filename(self):
        return self.filename


## -------------------------------------------------------------------------------------------------
    def __cb_button_ok(self):
        self.filename = self.folder_var.get() + os.sep + self.fname_var.get() + self.suffix_var.get()
        self.quit()


## -------------------------------------------------------------------------------------------------
    def __cb_button_cancel(self):
        self.filename = ''
        self.quit()


## -------------------------------------------------------------------------------------------------
    def __cb_button_folder(self):
        f = askdirectory(initialdir=self.folder_var.get(), mustexist=True)
        if f != (): self.folder_var.set(f)
 
        self.focus_force()





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUISubplotRoot(SciUIFrame): 
    """
    Root class for specialized frame classes that embedd a matplotlib figure with one subplot into a
    Tkinter frame. Not intended for direct reuse. Please use inherited classes SciUISubplot2D, 
    SciUISubplot3D instead. 
    """

    C_TYPE              = 'SciUI Subplot'

    C_BACKEND           = 'TkAgg'
    
    C_FIG_FACECOLOR     = 'white'

    C_AX_RECTANGLE      = [0.1, 0.1, 0.85, 0.85]
    C_AX_FRAME          = True
    C_AX_FACECOLOR      = 'white'

## -------------------------------------------------------------------------------------------------
    def create_subplot(self):
        """
        Internally used to create a suitable subplot. New subplot needs to be bound to self.ax. To be
        redefined.
        """

        self.ax = None


## -------------------------------------------------------------------------------------------------
    def init_popup_menu(self): 
        if self.popup_menu != None: return
        SciUIFrame.init_popup_menu(self)
        
        self.popup_menu = Menu(self.shared_db.root, tearoff=False)
        self.popup_menu.add_command(label='Undock', state='disabled') #, command=self.__cb_my_first_entry)
        self.popup_menu.add_separator()
        self.popup_menu.add_command(label='Save As...', command=self.cb_save_plot) 

        
## -------------------------------------------------------------------------------------------------
    def init_component(self):
        """
        Initialization of component-specific elements at instance creation time. To be redefined.
        Please call super().init_component() at beginning of your implementation.
        """

        SciUIFrame.init_component(self)

        if ( self.width_perc == 0.0 ) and ( self.height_perc == 0.0 ):
            raise Exception ('Subplots need to be sized explicitly!')

        if matplotlib.get_backend() != self.C_BACKEND:
            matplotlib.use(self.C_BACKEND)

        self.padx               = 0
        self.pady               = 0
        self.frame_visible      = True
        self.dpi                = 100               # Matplotlib default density
        self.figure             = None
        self.canvas             = None
        self.width_inch         = 0
        self.width_inch_bak     = 0
        self.height_inch        = 0
        self.height_inch_bak    = 0
        self.set_flush_events(True)
        
        
## -------------------------------------------------------------------------------------------------
    def determine_frame_size(self):
        SciUIFrame.determine_frame_size(self)
        
        # Determine plot size in inches
        self.width_inch     = self.width_pix / self.dpi
        self.height_inch    = self.height_pix / self.dpi

        
## -------------------------------------------------------------------------------------------------
    def set_flush_events(self, p_flush:bool):
        self.canvas_flush_events = p_flush


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        SciUIFrame.refresh(self, p_parent_frame)
        
        # 1 Refresh Matplotlib figure, canvas, ax
        if self.figure == None:
            self.figure = Figure(figsize=(self.width_inch, self.height_inch), dpi=self.dpi, facecolor=self.C_FIG_FACECOLOR)
            
            # 1.1 Embedd figure into matplotlib backend
            self.canvas = FigureCanvasTkAgg(self.figure, master=self.frame) 
            self.create_subplot()

            if self.popup_menu != None: self.canvas.get_tk_widget().bind("<Button-3>",self.cb_popup_menu)
            
        elif ( self.width_inch != self.width_inch_bak ) or ( self.height_inch != self.height_inch_bak ):
            self.figure.set_size_inches(self.width_inch, self.height_inch) 
            self.width_inch_bak     = self.width_inch
            self.height_inch_bak    = self.height_inch
        
        # 2 Refresh custom objects
        self.refresh_custom()
        
        # 3 Update view
        if not self.shared_db.redraw: return
        self.canvas.draw()
        if self.canvas_flush_events: self.canvas.flush_events()
        self.canvas.get_tk_widget().grid()        
        
        
## -------------------------------------------------------------------------------------------------
    def refresh_custom(self):    
        """
        Additional refresh activities. To be redefined.
        """

        self.log(self.C_LOG_TYPE_I, 'refresh custom objects')

      
## -------------------------------------------------------------------------------------------------
    def add_component(self, p_component): 
        """
        Adding further subcomponents is disabled here.
        """
        pass
    
        
## -------------------------------------------------------------------------------------------------
    def cb_save_plot(self):
        self.log(self.C_LOG_TYPE_I, 'Plot save dialog called from context menu')

        dlg = SciUISubplotSaveDLG(p_fname=self.frame_text, p_logging=self._level)
        dlg.start()
        f   = dlg.get_filename()
        if f != '':
            self.figure.savefig(f)
            self.log(self.C_LOG_TYPE_I, 'Plot saved to "' + f + '"')
        else:
            self.log(self.C_LOG_TYPE_I, 'Plot save dialog cancelled')

        dlg.destroy()


## -------------------------------------------------------------------------------------------------
    def __del__(self):
        if self.figure != None:
            if self.ax != None:
                self.figure.delaxes(self.ax) 
                self.ax = None
                
            self.figure = None
            
        try:
            SciUIFrame.__del__(self)
        except:
            pass
    
    



## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUISubplot2D(SciUISubplotRoot): 
    """
    Specialized frame class that embedds a 2D matplotlib figure with one subplot into a Tkinter 
    frame. A cross hair cursor functinality with mouse event handling can be switched on/off via
    class constant C_AX_CURSOR.  
    """

    C_TYPE              = 'SciUI 2D Subplot'
    C_AX_RECTANGLE      = [0.1, 0.1, 0.85, 0.85]
    C_AX_CURSOR         = False
    C_AX_CURSOR_COLOR   = 'blue'   

## -------------------------------------------------------------------------------------------------
    def create_subplot(self):
        self.ax =  self.figure.add_axes(self.C_AX_RECTANGLE, frame_on=self.C_AX_FRAME, fc=self.C_AX_FACECOLOR)      
        if self.C_AX_CURSOR: self.__init_cursor()


## -------------------------------------------------------------------------------------------------
    def __init_cursor(self):   
        self.cursor = SciUICursor(self.ax, useblit=True, color=self.C_AX_CURSOR_COLOR, linewidth=0.5, linestyle='--' )
        self.cursor.connect_event('button_press_event', self.cb_mbutton_pressed)
        self.cursor.connect_event('button_release_event', self.cb_mbutton_released)
        self.cursor.connect_event('motion_notify_event', self.cb_mouse_moved)
        self.cursor.set_event_status('button_press_event', True)

        
## -------------------------------------------------------------------------------------------------
    def cb_mbutton_pressed(self, event):
        self.log(self.C_LOG_TYPE_I, 'left mouse button pressed')
        self.cursor.set_event_status('motion_notify_event', True)
        self.cursor.set_event_status('button_release_event', True)  

        
## -------------------------------------------------------------------------------------------------
    def cb_mouse_moved(self, event):
        self.log(self.C_LOG_TYPE_I, 'mouse moved')

        
## -------------------------------------------------------------------------------------------------
    def cb_mbutton_released(self, event): 
        self.log(self.C_LOG_TYPE_I, 'left mouse button released')        
        self.cursor.set_event_status('motion_notify_event', False)
        self.cursor.set_event_status('button_release_event', False)
        




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUISubplot3D(SciUISubplotRoot): 
    """
    Specialized frame class that embedds a 3D matplotlib figure with one subplot into a Tkinter frame.  
    """

    C_TYPE              = 'SciUI 3D Subplot'
    C_AX_RECTANGLE      = [0.1, 0.1, 0.85, 0.85]

## -------------------------------------------------------------------------------------------------
    def create_subplot(self):
        self.ax = Axes3D(self.figure, auto_add_to_figure=False, rect=self.C_AX_RECTANGLE, proj_type='persp')
        self.figure.add_axes(self.ax)





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUIFrameParam(SciUIFrame):
    """
    Class for parameter/text frames.
    """ 

    C_TYPE          = 'SciUI Parameter Frame'
    C_NAME          = ''
    
    C_FONT_FAMILY   = 'Lucida Grande'
    C_FONT_SIZE     = 10    
 
## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.fontstyle = tkFont.Font(family=self.C_FONT_FAMILY, size=self.C_FONT_SIZE)
 
 
 


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUIScenario(SciUIFrame): 
    """
    Top level class for an entire SciUI scenario that can be registered by the SciUI application 
    class. SciUI scenarios are visible and chooseable if the switches C_RELEASED and C_VISIBLE are
    set to True.
    """
    
    C_TYPE          = 'SciUI Scenario'
    C_NAME          = '????'
    C_VERSION       = '0.0.0'
    C_RELEASED      = False
    C_VISIBLE       = False    
        
## -------------------------------------------------------------------------------------------------
    def __init__(self, p_shared_db:SciUISharedDB, p_logging=Log.C_LOG_ALL): 
        SciUIFrame.__init__(self, p_shared_db, p_row=0, p_col=0, p_width_perc=1, p_height_perc=1, p_visible=p_shared_db.window_fullscreen, p_logging=p_logging)

        if p_shared_db.window_fullscreen:
            self.frame_text = self.C_TYPE + ' "' + self.C_NAME + '"'

        self.padx = 3
        self.pady = 1
